# -*- coding: utf-8 -*-
"""
Created on Mon Nov  1 11:49:36 2021

@author:        
    Eva-Maria Grommes
    Sarah Glaubitz

Additional co-authors can be found here:
https://github.com/cire-thk/bifacialSimu    

"""
import pandas as pd
from pvlib import solarposition

lat =50.935
lon = 6.992
alitude = 53

times = pd.date_range('2021-09-23 00:00', freq='T', periods=23040, tz='Etc/GMT-2')
solpos = solarposition.get_solarposition(times, lat, lon, alitude)

zenith = solpos.zenith

print(zenith)

zenith.to_csv('minute_zenith.csv', sep=',')


# -*- coding: utf-8 -*-
"""
Created on Tue Oct  5 18:55:49 2021

@author:        
    Eva-Maria Grommes
    Sarah Glaubitz

Additional co-authors can be found here:
https://github.com/cire-thk/bifacialSimu    


Converting minute values of GHI, DNI and DHI into houly values
"""

import pandas as pd
import numpy as np
import datetime

#------------------------------------------------------------------------------
'Input parameters'
Dictionary = {
'startHour' : (2021, 9, 23, 0),                     # yy, mm, dd, hh
'number_of_days' : 16,                              # number of days to calculate hourly data for
'input_file' : '10minute_Flughafen.csv',       # inset name of input file (utf8 file is needed)
}
#------------------------------------------------------------------------------


df1 = pd.read_csv(Dictionary['input_file'], sep=';', header=0)  
print(df1)
loop_number = (Dictionary['number_of_days'] * 24 * 6)     # 24 hours per day and 6 10 minutes per hour

GHI_minute = []     # array to hold minute GHI values
DHI_minute = []     # array to hold minute DHI values
cd = []      # array to hold minute datetime

for i in range(loop_number):
     
    for j in range(10):
        
        k = i * 10 + j
        
        GHI = df1.iloc[i]['GHI'] 
        DHI = df1.iloc[i]['DHI'] 
        
        currentDate = datetime.datetime(Dictionary['startHour'][0], Dictionary['startHour'][1], Dictionary['startHour'][2], Dictionary['startHour'][3]) + pd.to_timedelta(k, unit='m') 
        
        GHI_minute.append(GHI)
        DHI_minute.append(DHI)
        cd.append(currentDate)          # append the currentDate to cd array
         

    
# create pandas dataframe to save the four arrays and give them headers
df2 = pd.DataFrame({'datetime':cd, 'GHI':GHI_minute, 'DHI':DHI_minute})
print(df2)

# save pandas dataframe into a csv file
df2.to_csv('minute_GHI_DHI.csv', sep=';', index=False)
        
        


# -*- coding: utf-8 -*-
"""
Created on Mon Jun  7 11:39:16 2021

@author:        
    Eva-Maria Grommes

Additional co-authors can be found here:
https://github.com/cire-thk/bifacialSimu    

name:
    BifacialSimu - calculationHandler

overview:
    Contains all own calculations of BifacialSimu which are needed for 
    the bifacial simulation of PV-Modules


"""

# from IPython import get_ipython
# get_ipython().magic('reset -sf')

import pandas as pd #pandas = can read .csv as input
import matplotlib.pyplot as plt #display shadows
import numpy as np
import os #to import directories
from bifacial_radiance import *
import datetime
from tqdm import tqdm
import math
import dateutil.tz
import GUI
# seaborn makes your plots look better
try:
    import seaborn as sns
    sns.set(rc={"figure.figsize": (12, 6)})
    sns.set_style("darkgrid") # Other available styles: whitegrid, dark, white, ticks
    sns.set_context("paper") # Scaling the axes, also available: talk, poster, 
    sns.set_palette("muted") 
    sns.set_color_codes()
except ImportError:
    print('We suggest you install seaborn using conda or pip and rerun this cell')

# Path handling
rootPath = rootPath = os.path.realpath("../../")

from BifacialSimu.Handler import BifacialSimu_radiationHandler 

# Include paths

#sys.path.append(rootPath + "/BifacialSimu/Handler")




# electric-calculation Klasse
    
    # Functions for One-Diode Model

class Electrical_simulation:
    """
    Class that simulates the eletrical output from the irradiance simulation
    Currently, electrical simulation works with only Viewfactors and Viewfactors/Raytracing combination 
    
    Methods
    -------
    build_simulationReport: build a final simulation report that contains data from viewfactors and raytracing
    simulate_oneDiode: Applies the one diode model for electrical simulation. Needs module front and rear parameters to work correctly.
    simulate_simpleBifacial: Simple electrical simulation mode that doesn't need rear module parameters. 
                Applies bifaciality factor to calculate rear efficiency and fill factor.
    """
    
    ##### Function to combine radiation reports from Viewfactors and Raytracing if needed
    def build_simulationReport(df_reportVF, df_reportRT, simulationDict, resultsPath):
                      
        if simulationDict['simulationMode'] == 1:
            df_report = pd.concat([df_reportVF, df_reportRT], axis=1)
        if simulationDict['simulationMode'] == 2:
            df_report = df_reportVF
        if simulationDict['simulationMode'] == 3:
            df_report = df_reportRT
        #df_report = df_report.reindex(sorted(df_report.columns), axis=1)
        df_report.to_csv(resultsPath + "radiation_qabs_results.csv")
        
        return df_report
    
    def simulate_oneDiode(moduleDict, simulationDict, df_reportVF, df_reportRT, df_report, df, resultsPath):
        """
        Applies the one diode model for bifacial electrical simulation. Needs module front and rear parameters to work correctly.
        Calculates bifacial gain through a seperate monofacial electrical simulation.

        Parameters
        ----------
        moduleDict: module Dictionary containing module data
        simulationDict: simulation Dictionary, which can be found in BifacialSimuu_main.py
        df_reportVF: Viewfactor simulation report
        df_reportRT: Raytracing simulation report
        df_report: Final simulation report, containing VF and RT data
        resultsPath: output filepath
        df: helper DataFrame containing temperature for electrical simulation
        """
        
        
        
        # Build a final simutlation report
        df_report = Electrical_simulation.build_simulationReport(df_reportVF, df_reportRT, simulationDict, resultsPath)
        
        ####################################################
        # Variables required for electrical simulation
        
        # P_bi: Output power of bifacial module for bifacial illumination (W)
        # I_sc_bi: Short-circuit current of bifacial module for bifacial illumination (A)
        # V_oc: Open-circuit voltage of bifacial module for bifacial illumination (V)
        # FF_bi: Fill factor of bifacial module for bifacial illumination (%)
        # G_r: Irradiance on the rear side of the module (W/m2)
        # G_f: Irradiance on the front side of the module (W/m2)
        # R_isc: Relative current gain (dimensionless)
        # I_sc_f: Short-circuit current measured for front side illumination of the module at STC (A)
        # x: Irradiance ratio (dimensionless)
        # V_oc_f: Open-circuit voltage measured for front side illumina- tion of module at STC (V)
        # V_oc_r: Open-circuit voltage measured for rear side illumina- tion of module at STC (V)
        # I_sc_r: Short-circuit current measured for rear side illumination of the module at STC (A)
        # FF_f: Fill factor measured for front side illumination of the module at STC (%)
        # FF_r: Fill factor measured for rear side illumination of the module (%)
        # pFF: Pseudo fill factor (FF of the module considering no series resistance effect) (%)

        ####################################################
        # Definition of simulation parameter
        V_mpp_f0 = moduleDict['V_mpp_f']
        V_mpp_r0 = moduleDict['V_mpp_r']
        
        I_mpp_f0 = moduleDict['I_mpp_f']
        I_mpp_r0 = moduleDict['I_mpp_r']
        
        I_sc_r0 = moduleDict['I_sc_r']
        I_sc_f0 = moduleDict['I_sc_f']
        
        V_oc_r0 = moduleDict['V_oc_r']
        V_oc_f0 = moduleDict['V_oc_f']
        
        #module = moduleParameter['module']
        #inverter = moduleParameter['inverter']
        
        P_mpp0 = moduleDict['P_mpp']
        V_mpp0 = V_mpp_f0
        
        T_koeff_P = moduleDict['T_koeff_P'] 
        T_koeff_I = moduleDict['T_koeff_I'] 
        T_koeff_V = moduleDict['T_koeff_V'] 
        T_amb = moduleDict['T_amb']
        
        q_stc_front = 1000  # [W/m^2] 
        q_stc_rear = 1000   # [W/m^2] 
        
        # Calculation of fill factor for STC conditions
        FF_f0 = (I_mpp_f0 * V_mpp_f0)/(I_sc_f0 * V_oc_f0) 
        FF_r0 = (I_mpp_r0 * V_mpp_r0)/(I_sc_r0 * V_oc_r0) 
        
        
        dpi = 300 #Quality for plot export
        ####################################################
        # Bifacial performance Calculation
        
        # # Fillfactor Calculation for front and back
        # FF_f = (V_mpp_f * I_mpp_f)/(V_oc_f * I_sc_f) # Fill factor measured for front side illumination of the module at STC [%/100]
        # print("Fill Factor front: " + str(FF_f))
        
        # FF_r = (V_mpp_r * I_mpp_r)/(V_oc_r * I_sc_r) # Fill factor measured for front back illumination of the module at STC [%/100]
        # print("Fill Factor back: " + str(FF_r))
        # print ("\n")
        
        # Set Energy to Zero       
        sum_energy_b = 0
        sum_power_b = 0
        
        # Array to hold other arrays -> average after for loop
        P_bi_hourly_arrays = []
        P_m_hourly_arrays = []
        
        df_report['timestamp'] = df_report.index
        df_report = df_report.reset_index()
        df_report['corrected_timestamp'] = pd.to_datetime(df_report['timestamp'])
        df_report['time'] = df_report['corrected_timestamp'].dt.strftime('%m_%d_%H')
        df_report = df_report.set_index('time')
        
        
        
        
        if simulationDict['simulationMode'] == 3:
            df = df.reset_index()
            
            df['time'] = df['corrected_timestamp'].dt.strftime('%m_%d_%H')
            df = df.set_index('time')
            df['timestamp'] = df['corrected_timestamp'].dt.strftime('%m-%d %H:%M%')
            df['timestamp'] = pd.to_datetime(df['timestamp'])  
            #df['timestamp'] = df['timestamp'].dt.tz_localize(None)
            df = df.set_index('timestamp')
            
            dtStart = datetime.datetime(simulationDict['startHour'][0], simulationDict['startHour'][1], simulationDict['startHour'][2], simulationDict['startHour'][3], tzinfo=dateutil.tz.tzoffset(None, simulationDict['utcOffset']*60*60))
        
        
            dtEnd = datetime.datetime(simulationDict['endHour'][0], simulationDict['endHour'][1], simulationDict['endHour'][2], simulationDict['endHour'][3], tzinfo=dateutil.tz.tzoffset(None, simulationDict['utcOffset']*60*60))
            mask = (df.index >= dtStart) & (df.index <= dtEnd) 
            df = df.loc[mask]
        
        df['time'] = df['corrected_timestamp'].dt.strftime('%m_%d_%H')
        df = df.set_index('time')
        
        
        # Loop to calculate the Bifacial Output power for every row in every hour
        for i in tqdm(range(0, simulationDict['nRows'])):
            
            key_front = "row_" + str(i) + "_qabs_front"
            key_back = "row_" + str(i) + "_qabs_back"
        
            P_bi_hourly = []
          
            for index, row in df_report.iterrows():
                
                row_qabs_front = df_report.loc[index,key_front]
                row_qabs_back = df_report.loc[index,key_back]
                T_Current = df.loc[index,'temperature']
                
                
                #print("front: " + str(row_qabs_front))
                #print("back: " + str(row_qabs_back))
                if math.isnan(row_qabs_front) or row_qabs_front < 0.0:
                    row_qabs_front = 0
                    
                if math.isnan(row_qabs_back) or row_qabs_back < 0.0:
                    row_qabs_back = 0

                
                if row_qabs_back + row_qabs_front > 0.0:
                    
                    I_sc_f = I_sc_f0 * (1 + T_koeff_I * (T_Current - T_amb))     
                    I_sc_r = I_sc_r0 * (1 + T_koeff_I * (T_Current - T_amb))
                    
                    V_oc_f = V_oc_f0 * (1 + T_koeff_V * (T_Current - T_amb))
                    V_oc_r = V_oc_r0 * (1 + T_koeff_V * (T_Current - T_amb))
                    
                    FF_f = FF_f0 * ((1 + T_koeff_P * (T_Current-T_amb)) / ((1 + T_koeff_I * (T_Current - T_amb)) * (1 + T_koeff_V * (T_Current - T_amb))))
                    FF_r = FF_r0 * ((1 + T_koeff_P * (T_Current-T_amb)) / ((1 + T_koeff_I * (T_Current - T_amb)) * (1 + T_koeff_V * (T_Current - T_amb))))
                    
                    I_sc_b = (row_qabs_front / q_stc_front) * I_sc_f + (row_qabs_back / q_stc_rear) * I_sc_r
                    R_I_sc_b = I_sc_b / I_sc_f
                    V_oc_b = V_oc_f + ((V_oc_r - V_oc_f) * np.log(R_I_sc_b) / np.log(I_sc_r / I_sc_f))
                    
                    pFF = ((I_sc_r0/I_sc_f0) * FF_f0 - (FF_r0 * (V_oc_r0 / V_oc_f0))) / ((I_sc_r0/I_sc_f0) - (V_oc_r0 / V_oc_f0))
                    FF_b = pFF - (R_I_sc_b * (V_oc_f0 / V_oc_b) * (pFF - FF_f0))
                
                    P_bi = FF_b * V_oc_b * I_sc_b
                    #print("Power: " + str(P_bi))
            
                    sum_energy_b += P_bi # Sum up the energy of every row in every hour
            
                else:
                    P_bi=0
                
                P_bi_hourly.append(P_bi)
                
            # Append P_bi_hourly array to arrays
            P_bi_hourly_arrays.append(P_bi_hourly)
                
                
        P_bi_hourly_average = []
        
        for i in tqdm(range(0, len(P_bi_hourly_arrays[0]))):
            sum = 0
          
            for j in range(0, len(P_bi_hourly_arrays)):
                sum += P_bi_hourly_arrays[j][i]
                
            average = sum / float(len(P_bi_hourly_arrays))
            
            P_bi_hourly_average.append(average)
            
       
        
        # The time gets implemented in the GUI
    # p_bi_df.to_csv(resultsPath + "electrical_simulation" + datetime.datetime.now().strftime("%Y-%m-%d-%H-%M") + ".csv")
        
        
        annual_power_per_module_b = (sum_energy_b/simulationDict['nRows']) #[W] annual bifacial output power per module
        '''print("Yearly bifacial output power per module: " + str(annual_power_per_module_b) + " W/module")
        print("Yearly bifacial output energy per module: " + str(annual_power_per_module_b/1000) + " kWh/module") # Because the input data is per hour, the Energy is equivalent to the performance
        print ("\n")'''
        
        annual_power_per_peak_b = (sum_energy_b/(P_mpp0 * simulationDict['nRows']))   #[W/Wp] annual bifacial output power per module peak power
        '''print("Yearly bifacial output power per module peak power: " + str(annual_power_per_peak_b) + " W/Wp")
        print("Yearly bifacial output energy per module peak power: " + str(annual_power_per_peak_b) + " kWh/kWp") # Because the input data is per hour, the Energy is equivalent to the performance 
        print ("\n")'''
        
        module_area = (simulationDict['moduley'] *simulationDict['nModsy'] *simulationDict['modulex'])
        
        annual_power_per_area_b = (annual_power_per_module_b / module_area)    #[W/m^2] annual bifacial poutput power per module area
        '''print("Yearly bifacial output power per module area: " + str(annual_power_per_area_b) + " W/m^2")
        print("Yearly bifacial output energy per module area: " + str(annual_power_per_area_b/1000) + " kWh/m^2") # Because the input data is per hour, the Energy is equivalent to the performance 
        print ("\n")'''
        
        # Plot total qinc front and back for every row
      
        
        f = plt.Figure(figsize=(12, 3))
        ax1 = f.subplots(1)
        ax1.locator_params(tight=True, nbins=6)
        ax1.plot(P_bi_hourly)
        ax1.set_title('Bifacial output Power hourly')
        ax1.set_xlabel('Hour')
        ax1.set_ylabel('W')
        f.savefig("P_bi_hourly" + datetime.datetime.now().strftime("%Y-%m-%d-%H-%M") + ".png", dpi = dpi)
        #plt.show()()
        #sns wird entfernt da die Plots in der GUI ansonsten nicht richtig angezeigt werden können außerhalb der Konsole
        ##plt.show()(sns)
        
        
        ####################################################
        # Monofacial performance Calculation
        
        # Set Energy to Zero       
        sum_energy_m = 0
        sum_power_m = 0
        
        # Loop to calculate the Monofacial Output power for every row in every hour
        for i in tqdm(range(0, simulationDict['nRows'])):
            
            key_front_mono = "row_" + str(i) + "_qabs_front"
            P_m_hourly = []
            
            for index, row in df_report.iterrows():
                
                #SG
                row_qabs_front = df_report.loc[index,key_front_mono]
                T_Current = df.loc[index,'temperature']

                if math.isnan(row_qabs_front):
                    row_qabs_front = 0 
                
                if  row_qabs_front > 0.0:
              
                    V_oc_f = V_oc_f0 * (1 + T_koeff_V * (T_Current - T_amb) + moduleDict['zeta'] * np.log(row_qabs_front / q_stc_front))
                    I_sc_f = I_sc_f0 * (1 + T_koeff_I * (T_Current - T_amb)) * (row_qabs_front / q_stc_front)
                    P_m = FF_f0 * V_oc_f * I_sc_f
                
                    #print("Power: " + str(P_bi))
             
                    sum_energy_m += P_m # Sum up the energy of every row in every hour
                else:
                    P_m = 0
                    
                P_m_hourly.append(P_m)
            
            # Append P_m_hourly array to arrays
            P_m_hourly_arrays.append(P_m_hourly)
        
        
        P_m_hourly_average = []
        
        for i in tqdm(range(0, len(P_m_hourly_arrays[0]))):
            sum = 0
          
            for j in range(0, len(P_m_hourly_arrays)):
                sum += P_m_hourly_arrays[j][i]
                
            average_m = sum / float(len(P_m_hourly_arrays))
            
            P_m_hourly_average.append(average_m)
                 #else:
                    #print("Power: 0.0")
        
        annual_power_per_module_m = (sum_energy_m/simulationDict['nRows']) #[W] annual monofacial output power per module
        '''print("Yearly monofacial output power per module: " + str(annual_power_per_module_m) + " W/module")
        print("Yearly monofacial output energy per module: " + str(annual_power_per_module_m/1000) + " kWh/module") # Because the input data is per hour, the Energy is equivalent to the performance
        print ("\n")'''
        
        annual_power_per_peak_m = (sum_energy_m/(P_mpp0 * simulationDict['nRows']))   #[W/Wp] annual monofacial output power per module peak power
        '''print("Yearly monofacial output power per module peak power: " + str(annual_power_per_peak_m) + " W/Wp")
        print("Yearly monofacial output energy per module peak power: " + str(annual_power_per_peak_m) + " kWh/kWp") # Because the input data is per hour, the Energy is equivalent to the performance 
        print ("\n")'''
        
        module_area = (simulationDict['moduley'] *simulationDict['nModsy']* simulationDict['modulex'])
        
        annual_power_per_area_m = (annual_power_per_module_m / module_area)    #[W/m^2] annual monofacial poutput power per module area
        '''print("Yearly monofacial output power per module area: " + str(annual_power_per_area_m) + " W/m^2")
        print("Yearly monofacial output energy per module area: " + str(annual_power_per_area_m/1000) + " kWh/m^2") # Because the input data is per hour, the Energy is equivalent to the performance 
        print ("\n")'''
        
        ####################################################
        
        # Bifacial Gain Calculation
        
        Bifacial_gain= (annual_power_per_peak_b - annual_power_per_peak_m) / annual_power_per_peak_m
        print("Bifacial Gain: " + str(Bifacial_gain*100) + " %")
        
                
        # Create dataframe with data
        p_bi_df = pd.DataFrame({"timestamps":df_report.index, "P_bi ": P_bi_hourly_average, "P_m ": P_m_hourly_average})
        p_bi_df.set_index("timestamps")
        p_bi_df.to_csv(resultsPath + "electrical_simulation" + datetime.datetime.now().strftime("%Y-%m-%d-%H-%M") + ".csv")
        
        #Plot for Bifacial Power Output + Bifacial Gain
        GUI.Window.makePlotBifacialRadiance(resultsPath,Bifacial_gain)
        
        return Bifacial_gain*100
        
    def simulate_simpleBifacial(moduleDict, simulationDict, df_reportVF, df_reportRT, df_report, df, resultsPath):
        """
        Applies a simplified version of the electrical simulation after PVSyst. Uses bifaciality factor to calculate rear efficiency and fill factors.
        Rear open-circuit voltage and short-circuit current are calculated using rear irradiance and temperature. 
        The rear output is then determined through the rear fill factor. Then the bifacial electrical output is calculated by adding front and rear output.
        Calculates bifacial gain through a seperate monofacial electrical simulation.

        Parameters
        ----------
        moduleDict: module Dictionary containing module data
        simulationDict: simulation Dictionary, which can be found in BifacialSimu_main.py
        df_reportVF: Viewfactor simulation report
        df_reportRT: Raytracing simulation report
        df_report: Final simulation report, containing VF and RT data
        resultsPath: output filepath
        df: helper DataFrame containing temperature for electrical simulation
        """
        
        
        # Build a final simulation report
        df_report = Electrical_simulation.build_simulationReport(df_reportVF, df_reportRT, simulationDict, resultsPath)
        
        ####################################################
        # Variables required for electrical simulation
        
        # P_bi: Output power of bifacial module for bifacial illumination (W)
        # I_sc_bi: Short-circuit current of bifacial module for bifacial illumination (A)
        # V_oc: Open-circuit voltage of bifacial module for bifacial illumination (V)
        # FF_bi: Fill factor of bifacial module for bifacial illumination (%)
        # G_r: Irradiance on the rear side of the module (W/m2)
        # G_f: Irradiance on the front side of the module (W/m2)
        # R_isc: Relative current gain (dimensionless)
        # I_sc_f: Short-circuit current measured for front side illumination of the module at STC (A)
        # x: Irradiance ratio (dimensionless)
        # V_oc_f: Open-circuit voltage measured for front side illumina- tion of module at STC (V)
        # V_oc_r: Open-circuit voltage measured for rear side illumina- tion of module at STC (V)
        # I_sc_r: Short-circuit current measured for rear side illumination of the module at STC (A)
        # FF_f: Fill factor measured for front side illumination of the module at STC (%)
        # FF_r: Fill factor measured for rear side illumination of the module (%)
        # pFF: Pseudo fill factor (FF of the module considering no series resistance effect) (%)

        ####################################################
        # Definition of simulation parameter if only front parameters are available
        # Procedure after PVSyst
        V_mpp_f0 = moduleDict['V_mpp_f']
            
        I_mpp_f0 = moduleDict['I_mpp_f']
                
        I_sc_f0 = moduleDict['I_sc_f']
        
        V_oc_f0 = moduleDict['V_oc_f']
        
        bi_factor = moduleDict['bi_factor']
        
        #module = moduleParameter['module']
        #inverter = moduleParameter['inverter']
        
        P_mpp0 = moduleDict['P_mpp']
        V_mpp0 = V_mpp_f0
        
        T_koeff_P = moduleDict['T_koeff_P'] 
        T_koeff_I = moduleDict['T_koeff_I'] 
        T_koeff_V = moduleDict['T_koeff_V'] 
        T_amb = moduleDict['T_amb']
        
        q_stc_front = 1000  # [W/m^2] 
        q_stc_rear = 1000   # [W/m^2] 
        
        # Calculation of fill factor for STC conditions
        FF_f0 = (I_mpp_f0 * V_mpp_f0)/(I_sc_f0 * V_oc_f0)
        
        #n_back = (moduleDict['bi_factor']*moduleDict['n_front'])
        
        #FF_fr = (n_back*q_stc_rear*(simulationDict['moduley'])*(simulationDict['modulex']))/(I_sc_f0 * V_oc_f0)
                       
        dpi = 150 #Quality for plot export
        
        # Set Energy to Zero       
        sum_energy_b = 0
        sum_power_b = 0
        
        # Array to hold other arrays -> average after for loop
        P_bi_hourly_arrays = []
        
        df_report['timestamp'] = df_report.index
        df_report = df_report.reset_index()
        df_report['corrected_timestamp'] = pd.to_datetime(df_report['timestamp'])
        df_report['time'] = df_report['corrected_timestamp'].dt.strftime('%m_%d_%H')
        df_report = df_report.set_index('time')
        
        if simulationDict['simulationMode'] == 3:
            df = df.reset_index()
            
            df['time'] = df['corrected_timestamp'].dt.strftime('%m_%d_%H')
            df = df.set_index('time')
            df['timestamp'] = df['corrected_timestamp'].dt.strftime('%m-%d %H:%M%')
            df['timestamp'] = pd.to_datetime(df['timestamp'])  
            #df['timestamp'] = df['timestamp'].dt.tz_localize(None)
            df = df.set_index('timestamp')
            
            dtStart = datetime.datetime(simulationDict['startHour'][0], simulationDict['startHour'][1], simulationDict['startHour'][2], simulationDict['startHour'][3], tzinfo=dateutil.tz.tzoffset(None, simulationDict['utcOffset']*60*60))
        
        
            dtEnd = datetime.datetime(simulationDict['endHour'][0], simulationDict['endHour'][1], simulationDict['endHour'][2], simulationDict['endHour'][3], tzinfo=dateutil.tz.tzoffset(None, simulationDict['utcOffset']*60*60))
            mask = (df.index >= dtStart) & (df.index <= dtEnd) 
            df = df.loc[mask]
        
        df['time'] = df['corrected_timestamp'].dt.strftime('%m_%d_%H')
        df = df.set_index('time')
        print(df_report)
        
        
        
        # Loop to calculate the Bifacial Output power for every row in every hour
        for i in tqdm(range(0, simulationDict['nRows'])):
            
            key_front = "row_" + str(i) + "_qabs_front"
            key_back = "row_" + str(i) + "_qabs_back"
        
            P_bi_hourly = []
          
            for index, row in df_report.iterrows():
                
                #row_qabs_front = row[key_front]
                #row_qabs_back = row[key_back]
                
                row_qabs_front = df_report.loc[index,key_front]
                row_qabs_back = df_report.loc[index,key_back]
                row_qabs_combined = row_qabs_front + (row_qabs_back*bi_factor)
                T_Current = df.loc[index,'temperature']
                
                
                # calculation of frontside power output
                if math.isnan(row_qabs_front) or row_qabs_front < 0.0:
                    row_qabs_front = 0
                    P_bi = 0     

                # calculation of backside power output
                elif math.isnan(row_qabs_back) or row_qabs_back < 0.0:
                    row_qabs_back = 0
                    P_bi = 0
               
                else:
                    V_oc_f = V_oc_f0 * (1 + T_koeff_V * (T_Current - T_amb) + moduleDict['zeta'] * np.log(row_qabs_combined / q_stc_front))
                    I_sc_f = I_sc_f0 * (1 + T_koeff_I * (T_Current - T_amb)) * (row_qabs_combined / q_stc_front)
                    P_bi = FF_f0 * V_oc_f * I_sc_f
                

                sum_energy_b += P_bi # Sum up the energy of every row in every hour

                P_bi_hourly.append(P_bi)
                
            # Append P_bi_hourly array to arrays
            P_bi_hourly_arrays.append(P_bi_hourly)

            print(sum_energy_b)
                
        P_bi_hourly_average = []
        
        for i in tqdm(range(0, len(P_bi_hourly_arrays[0]))):
            sum = 0
          
            for j in range(0, len(P_bi_hourly_arrays)):
                sum += P_bi_hourly_arrays[j][i]
                
            average = sum / float(len(P_bi_hourly_arrays))
            
            P_bi_hourly_average.append(average)
            
                
        # Create dataframe with average data
        p_bi_df = pd.DataFrame({"timestamps":df_report.index, "P_bi ": P_bi_hourly_average})
        p_bi_df.set_index("timestamps")
        p_bi_df.to_csv(resultsPath + "electrical_simulation" + datetime.datetime.now().strftime("%Y-%m-%d-%H-%M") + ".csv")
        
        
        annual_power_per_module_b = (sum_energy_b/simulationDict['nRows']) #[W] annual bifacial output power per module
        '''print("Yearly bifacial output power per module: " + str(annual_power_per_module_b) + " W/module")
        print("Yearly bifacial output energy per module: " + str(annual_power_per_module_b/1000) + " kWh/module") # Because the input data is per hour, the Energy is equivalent to the performance
        print ("\n")'''
        
        annual_power_per_peak_b = (sum_energy_b/(P_mpp0 * simulationDict['nRows']))   #[W/Wp] annual bifacial output power per module peak power
        '''print("Yearly bifacial output power per module peak power: " + str(annual_power_per_peak_b) + " W/Wp")
        print("Yearly bifacial output energy per module peak power: " + str(annual_power_per_peak_b) + " kWh/kWp") # Because the input data is per hour, the Energy is equivalent to the performance 
        print ("\n")'''
        
        module_area = (simulationDict['moduley'] * simulationDict['nModsy'] * simulationDict['modulex'])
        
        annual_power_per_area_b = (annual_power_per_module_b / module_area)    #[W/m^2] annual bifacial poutput power per module area
        '''print("Yearly bifacial output power per module area: " + str(annual_power_per_area_b) + " W/m^2")
        print("Yearly bifacial output energy per module area: " + str(annual_power_per_area_b/1000) + " kWh/m^2") # Because the input data is per hour, the Energy is equivalent to the performance 
        print ("\n")'''
        
        # Plot total qinc front and back for every row
        f = plt.Figure(figsize=(12, 3))
        ax1 = f.subplots(1)
        ax1.locator_params(tight=True, nbins=6)
        #f.plot(P_bi_hourly)
        ax1.set_title('Bifacial output Power hourly')
        ax1.set_xlabel('Hour')
        ax1.set_ylabel('W')
        f.savefig("P_bi_hourly" + datetime.datetime.now().strftime("%Y-%m-%d-%H-%M") + ".png", dpi = dpi)
        #plt.show()()
        ##plt.show()(sns)
         
        ####################################################
        # Monofacial performance Calculation
        
        # Set Energy to Zero       
        sum_energy_m = 0
        sum_power_m = 0
        
        # Loop to calculate the Monofacial Output power for every row in every hour
        for i in tqdm(range(0, simulationDict['nRows'])):
            
            key_front_mono = "row_" + str(i) + "_qabs_front"
            
            for index, row in df_report.iterrows():
                
                #SG
                #row_qabs_front = row[key_front_mono]
                #T_Current = df.loc[index,'temperature']
                row_qabs_front = df_report.loc[index,key_front_mono]
                T_Current = df.loc[index,'temperature']

                # calculation of frontside power output
                if math.isnan(row_qabs_front) or row_qabs_front < 0.0:
                    row_qabs_front = 0
                    P_m = 0 
                    
                else:
                    V_oc_f = V_oc_f0 * (1 + T_koeff_V * (T_Current - T_amb) + moduleDict['zeta'] * np.log(row_qabs_front / q_stc_front))
                    I_sc_f = I_sc_f0 * (1 + T_koeff_I * (T_Current - T_amb)) * (row_qabs_front / q_stc_front)
                    P_m = FF_f0 * V_oc_f * I_sc_f
               
                sum_energy_m += P_m # Sum up the energy of every row in every hour
                

        
        annual_power_per_module_m = (sum_energy_m/simulationDict['nRows']) #[W] annual monofacial output power per module
        '''print("Yearly monofacial output power per module: " + str(annual_power_per_module_m) + " W/module")
        print("Yearly monofacial output energy per module: " + str(annual_power_per_module_m/1000) + " kWh/module") # Because the input data is per hour, the Energy is equivalent to the performance
        print ("\n")'''
        
        annual_power_per_peak_m = (sum_energy_m/(P_mpp0 * simulationDict['nRows']))   #[W/Wp] annual monofacial output power per module peak power
        '''print("Yearly monofacial output power per module peak power: " + str(annual_power_per_peak_m) + " W/Wp")
        print("Yearly monofacial output energy per module peak power: " + str(annual_power_per_peak_m) + " kWh/kWp") # Because the input data is per hour, the Energy is equivalent to the performance 
        print ("\n")'''
        
        module_area = (simulationDict['moduley'] *simulationDict['nModsy']* simulationDict['modulex'])
        
        annual_power_per_area_m = (annual_power_per_module_m / module_area)    #[W/m^2] annual monofacial poutput power per module area
        '''print("Yearly monofacial output power per module area: " + str(annual_power_per_area_m) + " W/m^2")
        print("Yearly monofacial output energy per module area: " + str(annual_power_per_area_m/1000) + " kWh/m^2") # Because the input data is per hour, the Energy is equivalent to the performance 
        print ("\n")'''
        
        ####################################################
        
        # Bifacial Gain Calculation
        
        Bifacial_gain= (annual_power_per_peak_b - annual_power_per_peak_m) / annual_power_per_peak_m
        print("Bifacial Gain: " + str(Bifacial_gain*100) + " %")
        
        #Plot for Bifacial Power Output + Bifacial Gain
        GUI.Window.makePlotBifacialRadiance(resultsPath,Bifacial_gain)     
        
        return Bifacial_gain*100
        
        
    def simulate_doubleDiode(moduleDict, simulationDict, df_reportVF, df_reportRT, df_report, df, resultsPath):
        """
        Applies the one diode model for bifacial electrical simulation. Needs module front and rear parameters to work correctly.
        Calculates bifacial gain through a seperate monofacial electrical simulation.

        Parameters
        ----------
        moduleDict: module Dictionary containing module data
        simulationDict: simulation Dictionary, which can be found in BifacialSimuu_main.py
        df_reportVF: Viewfactor simulation report
        df_reportRT: Raytracing simulation report
        df_report: Final simulation report, containing VF and RT data
        resultsPath: output filepath
        df: helper DataFrame containing temperature for electrical simulation
        """
        
        
        
        # Build a final simutlation report
        df_report = Electrical_simulation.build_simulationReport(df_reportVF, df_reportRT, simulationDict, resultsPath)
        
        ####################################################
        # Variables required for electrical simulation
        
        # P_bi: Output power of bifacial module for bifacial illumination (W)
        # I_sc_bi: Short-circuit current of bifacial module for bifacial illumination (A)
        # V_oc: Open-circuit voltage of bifacial module for bifacial illumination (V)
        # FF_bi: Fill factor of bifacial module for bifacial illumination (%)
        # G_r: Irradiance on the rear side of the module (W/m2)
        # G_f: Irradiance on the front side of the module (W/m2)
        # R_isc: Relative current gain (dimensionless)
        # I_sc_f: Short-circuit current measured for front side illumination of the module at STC (A)
        # x: Irradiance ratio (dimensionless)
        # V_oc_f: Open-circuit voltage measured for front side illumina- tion of module at STC (V)
        # V_oc_r: Open-circuit voltage measured for rear side illumina- tion of module at STC (V)
        # I_sc_r: Short-circuit current measured for rear side illumination of the module at STC (A)
        # FF_f: Fill factor measured for front side illumination of the module at STC (%)
        # FF_r: Fill factor measured for rear side illumination of the module (%)
        # pFF: Pseudo fill factor (FF of the module considering no series resistance effect) (%)

        ####################################################
        # Definition of simulation parameter
        V_mpp_f0 = moduleDict['V_mpp_f']
        V_mpp_r0 = moduleDict['V_mpp_r']
        
        I_mpp_f0 = moduleDict['I_mpp_f']
        I_mpp_r0 = moduleDict['I_mpp_r']
        
        I_sc_r0 = moduleDict['I_sc_r']
        I_sc_f0 = moduleDict['I_sc_f']
        
        V_oc_r0 = moduleDict['V_oc_r']
        V_oc_f0 = moduleDict['V_oc_f']
        
        #module = moduleParameter['module']
        #inverter = moduleParameter['inverter']
        
        P_mpp0 = moduleDict['P_mpp']
        V_mpp0 = V_mpp_f0
        
        T_koeff_P = moduleDict['T_koeff_P'] 
        T_koeff_I = moduleDict['T_koeff_I'] 
        T_koeff_V = moduleDict['T_koeff_V'] 
        T_amb = moduleDict['T_amb']
        Ns = moduleDict['Ns']      #Number of cells in module
        
        k = 1.3806503 * 10**(-23)       #Boltzmann constant [J/K]
        q_ec = 1.60217646 * 10**(-19)   #electron charge [C]
        
        q_stc_front = 1000  # [W/m^2] 
        q_stc_rear = 1000   # [W/m^2] 
        
        # Calculation of fill factor for STC conditions
        FF_f0 = (I_mpp_f0 * V_mpp_f0)/(I_sc_f0 * V_oc_f0) 
        FF_r0 = (I_mpp_r0 * V_mpp_r0)/(I_sc_r0 * V_oc_r0) 
        
        
        dpi = 150 #Quality for plot export
        ####################################################
        # Bifacial performance Calculation
        
        # # Fillfactor Calculation for front and back
        # FF_f = (V_mpp_f * I_mpp_f)/(V_oc_f * I_sc_f) # Fill factor measured for front side illumination of the module at STC [%/100]
        # print("Fill Factor front: " + str(FF_f))
        
        # FF_r = (V_mpp_r * I_mpp_r)/(V_oc_r * I_sc_r) # Fill factor measured for front back illumination of the module at STC [%/100]
        # print("Fill Factor back: " + str(FF_r))
        # print ("\n")
        
        # Set Energy to Zero       
        sum_energy_b = 0
        sum_power_b = 0
        
        # Array to hold other arrays -> average after for loop
        P_bi_hourly_arrays = []
        P_m_hourly_arrays = []
        
        df_report['timestamp'] = df_report.index
        df_report = df_report.reset_index()
        df_report['corrected_timestamp'] = pd.to_datetime(df_report['timestamp'])
        df_report['time'] = df_report['corrected_timestamp'].dt.strftime('%m_%d_%H')
        df_report = df_report.set_index('time')
        
        
        
        
        if simulationDict['simulationMode'] == 3:
            df = df.reset_index()
            
            df['time'] = df['corrected_timestamp'].dt.strftime('%m_%d_%H')
            df = df.set_index('time')
            df['timestamp'] = df['corrected_timestamp'].dt.strftime('%m-%d %H:%M%')
            df['timestamp'] = pd.to_datetime(df['timestamp'])  
            #df['timestamp'] = df['timestamp'].dt.tz_localize(None)
            df = df.set_index('timestamp')
            
            dtStart = datetime.datetime(simulationDict['startHour'][0], simulationDict['startHour'][1], simulationDict['startHour'][2], simulationDict['startHour'][3], tzinfo=dateutil.tz.tzoffset(None, simulationDict['utcOffset']*60*60))
        
        
            dtEnd = datetime.datetime(simulationDict['endHour'][0], simulationDict['endHour'][1], simulationDict['endHour'][2], simulationDict['endHour'][3], tzinfo=dateutil.tz.tzoffset(None, simulationDict['utcOffset']*60*60))
            mask = (df.index >= dtStart) & (df.index <= dtEnd) 
            df = df.loc[mask]
        
        df['time'] = df['corrected_timestamp'].dt.strftime('%m_%d_%H')
        df = df.set_index('time')
        
        #Diode ideality factors. a1 has to be 1 while a2 is flexible but it has to be above 1.2
        a1 = 1      
        a2 = 1.3
        
        #Tolerances for current and power in Mpp calculation algorythm
        tol_I = 0.001 
        tol_P = 0.002
        
        #Calculation of Parameters that do not change in the loop
        P_mpp_f0 = V_mpp_f0 * I_mpp_f0                      #P_mpp is calculated from Mpp Voltage and Current
        Vt = (Ns * k * (25+273.15)) / q_ec                  #Thermal voltage at STC                                     
        I_0_f0 = (I_sc_f0) / (np.exp((V_oc_f0) / (Vt))-1)   #Calculation of saturation current. It is assumed that its the same for both diodes

        #Starting values fpr Rs and Rp
        Rs_f0 = 0
        Rp_min = (V_mpp_f0 / (I_sc_f0 - I_mpp_f0)) - ((V_oc_f0 - V_mpp_f0) / I_mpp_f0)
        Rp_f0 = Rp_min
        
        #Starting parameters for Current, Voltage, and Power
        I = 0
        V = 0
        P = 0
        P1 = 0
        P_mpp_fs0 = 0
        
        #Definition of lists for the I-V P-V plots
        Vfplt = []
        Ifplt = []
        Pfplt = []
               
        #Rs and Rp calculation for the front side. They are calculated at STC. The algorythm tries to match the Mpp of the Module parameters
        #by adjusting Rs and calculating Rp based on Rs.
        for xf in range (1000000):
           
            
           #If the calculation takes too many iterations, an error message will show up.
           if xf == 999999:                     
                print ('Error: The calculation of the front resistances exeeded one million iterations. Please check your parameters.')
                     
           #'Calculation' of the photo current. At STC the I_sc is the photo current.        
           I_ph_f0 = I_sc_f0                                             
       
           #Newtons method to find the current that solves the equation f_I  
           f_I = I_ph_f0 - I_0_f0 * (np.exp((V+ I * Rs_f0) / Vt) + np.exp((V + I * Rs_f0) / (Vt * a2)) - 2) - ((V + I * Rs_f0) / Rp_f0) - I
           f_dI = (-1) * ((I_0_f0 * Rs_f0) / Vt) * np.exp((V + I * Rs_f0) / Vt) - ((I_0_f0 * Rs_f0) / (Vt * a2)) * np.exp((V+ I * Rs_f0) / (Vt * a2)) - (Rs_f0 / Rp_f0) - 1
       
           I2 = I - (f_I/f_dI)
              
           #Check if the new value of I2 is close enough (in tolerance) to the old value I
           if I + tol_I >= I2 and I2 >= I - tol_I:
               P2 = V * I2          #If the current was calculated correctly the power will be calculated
               
               Vfplt.append(V)      #The values of current, voltage and power are added to their designated lists
               Ifplt.append(I2)
               Pfplt.append(P2)
               
               if P2 > P1:          #Check if the new power value is greater than the last. If yes it becomes the new reference power value.
                   P1 = P2
               
               else:
                   P_mpp_fs0 = P1   #If the power does not increase however, the Mpp of this P-V curve is found
                   
                   
                           
               if V >= V_oc_f0:     # When V reaches V_oc the P-V curve is complete and it has to be checked if Mpp converged with the Module parameters
               
                   #If the calculated Mpp is within the tolerance the script tells the user the Values of Rs and Rp aswell as the amount of iterations
                   #it took to calculate them                   
                   if P_mpp_f0 + tol_P >= P_mpp_fs0 and P_mpp_fs0 >= P_mpp_f0 - tol_P:
                       print ('Front resistance calculation completed in ',xf,' iterations.', 'Rs_f0 =', Rs_f0, 'Rp_f0 =', Rp_f0)
                       
                       
                       #The P-V and I-V curve is plotted with the values since the last Rs, Rp adjustment
                       f = plt.Figure(figsize=(6, 6))
                       ax1 = f.subplots(1)
                       ax1.locator_params(tight=True, nbins=6)
                       ax1.plot(Vfplt[-round(V_oc_f0 * 10):], Ifplt[-round(V_oc_f0 * 10):], color = "red")
                       ax1.set_title('P-V I-V Curve front side', fontsize=14)
                       ax1.set_xlabel('Voltage [V]', fontsize=14)
                       ax1.set_ylabel('Current [I]', fontsize=14, color = 'red')
                       ax2=ax1.twinx()
                       ax2.plot(Vfplt[-round(V_oc_f0 * 10):], Pfplt[-round(V_oc_f0 * 10):], color = "blue")
                       ax2.set_ylabel('Power [W]', fontsize=14, color = 'blue')
                       f.savefig("P-V_I-V_Curve_front" + datetime.datetime.now().strftime("%Y-%m-%d-%H-%M") + ".png", dpi = dpi)
                       #plt.show()()
 
                       #Rs and Rp have been calculated so the script quits the loop 
                       break
                   
                   #If the calculated Mpp is not within tolerance, the reference power is reset and Rs gets increased by 10 mOhm 
                   else:
                       P1 = 0
                   
                       Rs_f0 = Rs_f0 + 0.01
                   
                       #Calculation of Rp with new Rs. For a better overview it is split into smaller calculations
                       r1 = V_mpp_f0 + I_mpp_f0 * Rs_f0
                       r2 = I_ph_f0
                       r3 = np.exp(r1/Vt)
                       r4 = np.exp(r1/(Vt*a2))
                       r5 = P_mpp_f0/V_mpp_f0
                       r6 = (r3 + r4 + 2)
                       r7 = r6 * I_0_f0

                       Rp_f0 = r1 / (r2 - r7 - r5)
                                     
                   V = -0.1     #V reached its max so it is set to -0.1 so in the first iteration we have I=0 and V=0
               I = 0
               V = V + 0.1
           else:
               I = I2           #If I2 was not close enough to I, I becomes I2 (newtons method)
        

        #After the calculation of Rs and Rp for the front side, the rear side is calculated in the same way, with values from the back side
        P_mpp_r0 = V_mpp_r0 * I_mpp_r0
        Vt = (Ns * k * (25+273.15)) / q_ec
        I_0_r0 = (I_sc_r0) / (np.exp((V_oc_r0) / (Vt))-1)   

        #Starting values fpr Rs and Rp
        Rs_r0 = 0
        Rp_min = (V_mpp_r0 / (I_sc_r0 - I_mpp_r0)) - ((V_oc_r0 - V_mpp_r0) / I_mpp_r0)
        Rp_r0 = Rp_min

        I = 0
        V = 0
        P = 0
        P1 = 0
        P_mpp_rs0 = 0
        
        Vrplt = []
        Irplt = []
        Prplt = []        
        
        #Rs and Rp calculation for the rear side. They are calculated at STC. The algorythm tries to match the Mpp of the Module parameters
        #by adjusting Rs and calculating Rp based on Rs.
        for xr in range (1000000):
       
           if xr == 999999:
                print ('Error: The calculation of the rear resistances exeeded one million iterations. Please check your parameters.')
        
           
           
                  
           #'Calculation' of the photo current. At STC the I_sc is the photo current.        
           I_ph_r0 = I_sc_r0                                             
       
           #Newtons method to find the current that solves the equation f_I  
           f_I = I_ph_r0 - I_0_r0 * (np.exp((V+ I * Rs_r0) / Vt) + np.exp((V + I * Rs_r0) / (Vt * a2)) - 2) - ((V + I * Rs_r0) / Rp_r0) - I
           f_dI = (-1) * ((I_0_r0 * Rs_r0) / Vt) * np.exp((V + I * Rs_r0) / Vt) - ((I_0_r0 * Rs_r0) / (Vt * a2)) * np.exp((V+ I * Rs_r0) / (Vt * a2)) - (Rs_r0 / Rp_r0) - 1
       
           I2 = I - (f_I/f_dI)
              
       
           if I + tol_I >= I2 and I2 >= I - tol_I:
               P2 = V * I2
               Vrplt.append(V)
               Irplt.append(I2)
               Prplt.append(P2)
               
               if P2 > P1:
                   P1 = P2
               
               else:
                   P_mpp_rs0 = P1              
                          
               if V >= V_oc_r0:
                   if P_mpp_r0 + tol_P >= P_mpp_rs0 and P_mpp_rs0 >= P_mpp_r0 - tol_P:
                       print ('Rear resistance calculation completed in ',xr,' iterations.', 'Rs_r0 =', Rs_r0, 'Rp_r0 =', Rp_r0)
                       
                       f, (ax1) = plt.subplots(1, figsize=(6, 6))
                       ax1.locator_params(tight=True, nbins=6)
                       ax1.plot(Vfplt[-round(V_oc_r0 * 10):], Ifplt[-round(V_oc_r0 * 10):], color = "red")
                       ax1.set_title('P-V I-V Curve back side', fontsize=14)
                       ax1.set_xlabel('Voltage [V]', fontsize=14)
                       ax1.set_ylabel('Current [I]', fontsize=14, color = 'red')
                       ax2=ax1.twinx()
                       ax2.plot(Vrplt[-round(V_oc_r0 * 10):], Prplt[-round(V_oc_r0 * 10):], color = "blue")
                       ax2.set_ylabel('Power [W]', fontsize=14, color = 'blue')
                       f.savefig("P-V_I-V_Curve_back" + datetime.datetime.now().strftime("%Y-%m-%d-%H-%M") + ".png", dpi = dpi)
                       #plt.show()()                       
                       
                       
                       break
                   else:
                       P1 = 0
                   
                       Rs_r0 = Rs_r0 + 0.01
                   
                       r1 = V_mpp_r0 + I_mpp_r0 * Rs_r0
                       r2 = I_ph_r0
                       r3 = np.exp(r1/Vt)
                       r4 = np.exp(r1/(Vt*a2))
                       r5 = P_mpp_r0/V_mpp_r0
                       r6 = (r3 + r4 + 2)
                       r7 = r6 * I_0_r0

                       Rp_r0 = r1 / (r2 - r7 - r5)
                                     
                   V = -0.1
               I = 0
               V = V + 0.1
           else:
               I = I2        
        

        
        sum_energy_m = 0
        sum_power_m = 0
        
        
        # Loop to calculate the Bifacial Output power for every row in every hour
        for i in tqdm(range(0, simulationDict['nRows'])):
            
            key_front = "row_" + str(i) + "_qabs_front"
            key_back = "row_" + str(i) + "_qabs_back"
            
            P_m_hourly = []
            P_bi_hourly = []
          
            for index, row in df_report.iterrows():
                
                row_qabs_front = df_report.loc[index,key_front]
                row_qabs_back = df_report.loc[index,key_back]
                T_Current = df.loc[index,'temperature']
                
                
                #print("front: " + str(row_qabs_front))
                #print("back: " + str(row_qabs_back))
                if math.isnan(row_qabs_front) or row_qabs_front < 0.0:
                    row_qabs_front = 0
                    
                if math.isnan(row_qabs_back) or row_qabs_back < 0.0:
                    row_qabs_back = 0

                
                if row_qabs_back + row_qabs_front > 0.0:
                    
                    #Now that Rs and Rp are calculated for both sides of the module, the power calculation starts.
                    #Values are now adjusted for temperature and later also irradiation
                    
                    I_sc_f = I_sc_f0 * (1 + T_koeff_I * (T_Current - T_amb))     
                    I_sc_r = I_sc_r0 * (1 + T_koeff_I * (T_Current - T_amb))
                    
                    V_oc_f = V_oc_f0 * (1 + T_koeff_V * (T_Current - T_amb))
                    V_oc_r = V_oc_r0 * (1 + T_koeff_V * (T_Current - T_amb))
                    
                    #setting starting parameters for the loop
                    I = 0
                    V = 0
                    P = 0
                    P1 = 0
                    P_mpp_sf = 0
                    
                    
                    #Modified version of the Rs, Rp calculation loop. Since Rs and Rp are now given, the loop needs only a fraction of iterations
                    #to calculate the power from a given irradiance and temperature
                    #The way this works is that the algorythm searches for the Mpp of the module with the given irradiance and temperature
                    #Just like a real PV system would do
                    #It 'draws' the P-V curve and finds the Mpp which is then the power output of the module
                    for yf in range (100000):
                        
                        if row_qabs_front == 0:
                            P_m = 0
                            P_mpp_sf = 0
                            break
                        
                        #Calculation of the photo current + correction for irrandiance and temperature
                        Vt_f = (Ns * k * (T_Current + 273.15)) / q_ec

                        #Calculation of the saturation current
                        I_0_f0 = (I_sc_f) / (np.exp((V_oc_f) / (Vt_f))-1)

                        #adjustment ot the photo current for irradiation
                        I_ph_f0 = I_sc_f                                             
                        I_ph_f = I_ph_f0 * (row_qabs_front / q_stc_front) 
              
                        #newthons method to find the matching current for a given voltage
                        f_I = I_ph_f - I_0_f0 * (np.exp((V+ I * Rs_f0) / Vt_f) + np.exp((V + I * Rs_f0) / (Vt_f * a2)) - 2) - ((V + I * Rs_f0) / Rp_f0) - I
                        f_dI = (-1) * ((I_0_f0 * Rs_f0) / Vt_f) * np.exp((V + I * Rs_f0) / Vt_f) - ((I_0_f0 * Rs_f0) / (Vt_f * a2)) * np.exp((V+ I * Rs_f0) / (Vt_f * a2)) - (Rs_f0 / Rp_f0) - 1
       
                        I2 = I - (f_I/f_dI)
              
       
                        if I + tol_I >= I2 and I2 >= I - tol_I:         #Once I is found, the Power check starts just like in xf loop
                            P2 = V * I2

                            if P2 > P1:                 #Check if the new power is higher than the last
                                P1 = P2                 #If this is true, it becomes the new reference value
               
                            else:
                                P_mpp_sf = P1           #The highest calculated power gets added to P_mpp_sf
                                
                            if V >= V_oc_f:             # If V reached V_oc, the P-V curve is complete and the Mpp can be searched
                                P_f = P_mpp_sf          #P_mpp_sf is the calculated Mpp for the Module for the given irradiance and and temperature
                                P_m = P_f
                                sum_energy_m += P_m     #The value gets added to a list for Bifacial gain calculation
                                yf = 0                  #Iterations get reset after successful Mpp calculation
                                break                   #The Mpp was found, the script quits the loop
                          
               
                            I = 0                       #Since there is only one P-V curve to calculate, V does not have to be reset
                            V = V + 0.1
                        else:
                            I = I2
                        
                            
                    ##################################
                    ###Same procedure for rear side###
                    ##################################
                    I = 0
                    V = 0
                    P = 0
                    P1 = 0
                    
                    
                    
                    for yr in range (100000):
                        
                        if row_qabs_front == 0:
                            P_mpp_sr = 0
                            break
                  
                        #Calculation of the photo current + correction for irrandiance and temperature
                        Vt_r = (Ns * k * (T_Current + 273.15)) / q_ec
 
                        I_0_r0 = (I_sc_r) / (np.exp((V_oc_r) / (Vt_r))-1)
       
                        I_ph_r0 = I_sc_r                                             
                        I_ph_r = I_ph_r0 * (row_qabs_back / q_stc_rear) 
              
                        f_I = I_ph_r - I_0_r0 * (np.exp((V+ I * Rs_r0) / Vt_r) + np.exp((V + I * Rs_r0) / (Vt_r * a2)) - 2) - ((V + I * Rs_r0) / Rp_r0) - I
                        f_dI = (-1) * ((I_0_r0 * Rs_r0) / Vt_r) * np.exp((V + I * Rs_r0) / Vt_r) - ((I_0_r0 * Rs_r0) / (Vt_r * a2)) * np.exp((V+ I * Rs_r0) / (Vt_r * a2)) - (Rs_r0 / Rp_r0) - 1
       
                        I2 = I - (f_I/f_dI)
              
       
                        if I + tol_I >= I2 and I2 >= I - tol_I:
                            P2 = V * I2

                            if P2 > P1:
                                P1 = P2
               
                            else:
                                P_mpp_sr = P1
                          
                            if V >= V_oc_r:
                                P_r = P_mpp_sr 
                                yr = 0
                                break
                          
               
                            I = 0
                            V = V + 0.1
                        else:
                            I = I2                    
                    
                
                    P_bi = P_mpp_sf + P_mpp_sr
                    #print("Power: " + str(P_bi))
            
                    sum_energy_b += P_bi # Sum up the energy of every row in every hour
            
                else:
                    P_m=0
                    P_bi=0
                
                P_m_hourly.append(P_m)
                P_bi_hourly.append(P_bi)
                
            # Append P_bi_hourly array to arrays
            P_m_hourly_arrays.append(P_m_hourly)
            P_bi_hourly_arrays.append(P_bi_hourly)
            
        P_bi_hourly_average = []
        
        for i in tqdm(range(0, len(P_bi_hourly_arrays[0]))):
            sum = 0
          
            for j in range(0, len(P_bi_hourly_arrays)):
                sum += P_bi_hourly_arrays[j][i]
                
            average = sum / float(len(P_bi_hourly_arrays))
            
            P_bi_hourly_average.append(average)
            
       
        
        # The time gets implemented in the GUI
    # p_bi_df.to_csv(resultsPath + "simulation" + datetime.datetime.now().strftime("%Y-%m-%d-%H-%M") + ".csv")
        
        
        annual_power_per_module_b = (sum_energy_b/simulationDict['nRows']) #[W] annual bifacial output power per module
        '''print("Yearly bifacial output power per module: " + str(annual_power_per_module_b) + " W/module")
        print("Yearly bifacial output energy per module: " + str(annual_power_per_module_b/1000) + " kWh/module") # Because the input data is per hour, the Energy is equivalent to the performance
        print ("\n")'''
        
        annual_power_per_peak_b = (sum_energy_b/(P_mpp0 * simulationDict['nRows']))   #[W/Wp] annual bifacial output power per module peak power
        '''print("Yearly bifacial output power per module peak power: " + str(annual_power_per_peak_b) + " W/Wp")
        print("Yearly bifacial output energy per module peak power: " + str(annual_power_per_peak_b) + " kWh/kWp") # Because the input data is per hour, the Energy is equivalent to the performance 
        print ("\n")'''
        
        module_area = (simulationDict['moduley'] *simulationDict['nModsy'] *simulationDict['modulex'])
        
        annual_power_per_area_b = (annual_power_per_module_b / module_area)    #[W/m^2] annual bifacial poutput power per module area
        '''print("Yearly bifacial output power per module area: " + str(annual_power_per_area_b) + " W/m^2")
        print("Yearly bifacial output energy per module area: " + str(annual_power_per_area_b/1000) + " kWh/m^2") # Because the input data is per hour, the Energy is equivalent to the performance 
        print ("\n")'''
        
        # Plot total qinc front and back for every row
      
        
        f = plt.Figure(figsize=(12, 3))
        ax1 = f.subplots(1)
        ax1.locator_params(tight=True, nbins=6)
        f.plot(P_bi_hourly)
        ax1.set_title('Bifacial output Power hourly')
        ax1.set_xlabel('Hour')
        ax1.set_ylabel('W')
        f.savefig("P_bi_hourly" + datetime.datetime.now().strftime("%Y-%m-%d-%H-%M") + ".png", dpi = dpi)
        #plt.show()()
        ##plt.show()(sns)
        
        
        ####################################################
        # Monofacial performance Calculation
        
        # Set Energy to Zero       
        #sum_energy_m = 0
        #sum_power_m = 0
        
        # Loop to calculate the Monofacial Output power for every row in every hour
        #for i in tqdm(range(0, simulationDict['nRows'])):
            
        #    key_front_mono = "row_" + str(i) + "_qabs_front"
        #    P_m_hourly = []
            
        #    for index, row in df_report.iterrows():
                
        #        #SG
        #        row_qabs_front = df_report.loc[index,key_front_mono]
        #        T_Current = df.loc[index,'temperature']

        #        if math.isnan(row_qabs_front):
        #            row_qabs_front = 0 
                
        #        if  row_qabs_front > 0.0:
              
        #            V_oc_f = V_oc_f0 * (1 + T_koeff_V * (T_Current - T_amb) + moduleDict['zeta'] * np.log(row_qabs_front / q_stc_front))
        #            I_sc_f = I_sc_f0 * (1 + T_koeff_I * (T_Current - T_amb)) * (row_qabs_front / q_stc_front)
        #            P_m = FF_f0 * V_oc_f * I_sc_f
                
                    #print("Power: " + str(P_bi))
             
        #            sum_energy_m += P_m # Sum up the energy of every row in every hour
        #        else:
        #            P_m = 0
                    
        #        P_m_hourly.append(P_m)
            
            # Append P_m_hourly array to arrays
        #    P_m_hourly_arrays.append(P_m_hourly)
        
        
        P_m_hourly_average = []
        
        for i in tqdm(range(0, len(P_m_hourly_arrays[0]))):
            sum = 0
          
            for j in range(0, len(P_m_hourly_arrays)):
                sum += P_m_hourly_arrays[j][i]
                
            average_m = sum / float(len(P_m_hourly_arrays))
            
            P_m_hourly_average.append(average_m)
                 #else:
                    #print("Power: 0.0")
        
        annual_power_per_module_m = (sum_energy_m/simulationDict['nRows']) #[W] annual monofacial output power per module
        '''print("Yearly monofacial output power per module: " + str(annual_power_per_module_m) + " W/module")
        print("Yearly monofacial output energy per module: " + str(annual_power_per_module_m/1000) + " kWh/module") # Because the input data is per hour, the Energy is equivalent to the performance
        print ("\n")'''
        
        annual_power_per_peak_m = (sum_energy_m/(P_mpp0 * simulationDict['nRows']))   #[W/Wp] annual monofacial output power per module peak power
        '''print("Yearly monofacial output power per module peak power: " + str(annual_power_per_peak_m) + " W/Wp")
        print("Yearly monofacial output energy per module peak power: " + str(annual_power_per_peak_m) + " kWh/kWp") # Because the input data is per hour, the Energy is equivalent to the performance 
        print ("\n")'''
        
        module_area = (simulationDict['moduley'] *simulationDict['nModsy']* simulationDict['modulex'])
        
        annual_power_per_area_m = (annual_power_per_module_m / module_area)    #[W/m^2] annual monofacial poutput power per module area
        '''print("Yearly monofacial output power per module area: " + str(annual_power_per_area_m) + " W/m^2")
        print("Yearly monofacial output energy per module area: " + str(annual_power_per_area_m/1000) + " kWh/m^2") # Because the input data is per hour, the Energy is equivalent to the performance 
        print ("\n")'''
        
        ####################################################
        
        # Bifacial Gain Calculation
        
        Bifacial_gain= (annual_power_per_peak_b - annual_power_per_peak_m) / annual_power_per_peak_m
        print("Bifacial Gain: " + str(Bifacial_gain*100) + " %")
        
                
        # Create dataframe with data
        p_bi_df = pd.DataFrame({"timestamps":df_report.index, "P_bi ": P_bi_hourly_average, "P_m ": P_m_hourly_average})
        p_bi_df.set_index("timestamps")
        p_bi_df.to_csv(resultsPath + "electrical_simulation" + datetime.datetime.now().strftime("%Y-%m-%d-%H-%M") + ".csv")
        #p_bi_df.to_csv(resultsPath + "electrical_simulation.csv")
        
        #Plot for Bifacial Power Output + Bifacial Gain
        GUI.Window.makePlotBifacialRadiance(resultsPath,Bifacial_gain)   
    
        return Bifacial_gain*100

    def simulate_doubleDiodeBi(moduleDict, simulationDict, df_reportVF, df_reportRT, df_report, df, resultsPath):
        """
        Applies the one diode model for bifacial electrical simulation. Needs module front and rear parameters to work correctly.
        Calculates bifacial gain through a seperate monofacial electrical simulation.

        Parameters
        ----------
        moduleDict: module Dictionary containing module data
        simulationDict: simulation Dictionary, which can be found in BifacialSimuu_main.py
        df_reportVF: Viewfactor simulation report
        df_reportRT: Raytracing simulation report
        df_report: Final simulation report, containing VF and RT data
        resultsPath: output filepath
        df: helper DataFrame containing temperature for electrical simulation
        """
        
        
        
        # Build a final simutlation report
        df_report = Electrical_simulation.build_simulationReport(df_reportVF, df_reportRT, simulationDict, resultsPath)
        
        ####################################################
        # Variables required for electrical simulation
        
        # P_bi: Output power of bifacial module for bifacial illumination (W)
        # I_sc_bi: Short-circuit current of bifacial module for bifacial illumination (A)
        # V_oc: Open-circuit voltage of bifacial module for bifacial illumination (V)
        # FF_bi: Fill factor of bifacial module for bifacial illumination (%)
        # G_r: Irradiance on the rear side of the module (W/m2)
        # G_f: Irradiance on the front side of the module (W/m2)
        # R_isc: Relative current gain (dimensionless)
        # I_sc_f: Short-circuit current measured for front side illumination of the module at STC (A)
        # x: Irradiance ratio (dimensionless)
        # V_oc_f: Open-circuit voltage measured for front side illumina- tion of module at STC (V)
        # V_oc_r: Open-circuit voltage measured for rear side illumina- tion of module at STC (V)
        # I_sc_r: Short-circuit current measured for rear side illumination of the module at STC (A)
        # FF_f: Fill factor measured for front side illumination of the module at STC (%)
        # FF_r: Fill factor measured for rear side illumination of the module (%)
        # pFF: Pseudo fill factor (FF of the module considering no series resistance effect) (%)

        ####################################################
        # Definition of simulation parameter
        V_mpp_f0 = moduleDict['V_mpp_f']
        V_mpp_r0 = moduleDict['V_mpp_r']
        
        I_mpp_f0 = moduleDict['I_mpp_f']
        I_mpp_r0 = moduleDict['I_mpp_r']
        
        I_sc_r0 = moduleDict['I_sc_r']
        I_sc_f0 = moduleDict['I_sc_f']
        
        V_oc_r0 = moduleDict['V_oc_r']
        V_oc_f0 = moduleDict['V_oc_f']
        
        #module = moduleParameter['module']
        #inverter = moduleParameter['inverter']
        
        P_mpp0 = moduleDict['P_mpp']
        V_mpp0 = V_mpp_f0
        
        T_koeff_P = moduleDict['T_koeff_P'] 
        T_koeff_I = moduleDict['T_koeff_I'] 
        T_koeff_V = moduleDict['T_koeff_V'] 
        T_amb = moduleDict['T_amb']
        
        # Note for later!:
        # moduleDict['Ns'] is not defined in moduleDict! (This can gave an error) 
        Ns = moduleDict['Ns']      #Number of cells in module
        
        k = 1.3806503 * 10**(-23)       #Boltzmann constant [J/K]
        q_ec = 1.60217646 * 10**(-19)   #electron charge [C]
        
        q_stc_front = 1000  # [W/m^2] 
        q_stc_rear = 1000   # [W/m^2] 
        
        # Calculation of fill factor for STC conditions
        FF_f0 = (I_mpp_f0 * V_mpp_f0)/(I_sc_f0 * V_oc_f0) 
        FF_r0 = (I_mpp_r0 * V_mpp_r0)/(I_sc_r0 * V_oc_r0) 
        
        
        dpi = 150 #Quality for plot export
        ####################################################
        # Bifacial performance Calculation
        
        # # Fillfactor Calculation for front and back
        # FF_f = (V_mpp_f * I_mpp_f)/(V_oc_f * I_sc_f) # Fill factor measured for front side illumination of the module at STC [%/100]
        # print("Fill Factor front: " + str(FF_f))
        
        # FF_r = (V_mpp_r * I_mpp_r)/(V_oc_r * I_sc_r) # Fill factor measured for front back illumination of the module at STC [%/100]
        # print("Fill Factor back: " + str(FF_r))
        # print ("\n")
        
        # Set Energy to Zero       
        sum_energy_b = 0
        sum_power_b = 0
        
        # Array to hold other arrays -> average after for loop
        P_bi_hourly_arrays = []
        P_m_hourly_arrays = []
        
        df_report['timestamp'] = df_report.index
        df_report = df_report.reset_index()
        df_report['corrected_timestamp'] = pd.to_datetime(df_report['timestamp'])
        df_report['time'] = df_report['corrected_timestamp'].dt.strftime('%m_%d_%H')
        df_report = df_report.set_index('time')
        
        
        
        
        if simulationDict['simulationMode'] == 3:
            df = df.reset_index()
            
            df['time'] = df['corrected_timestamp'].dt.strftime('%m_%d_%H')
            df = df.set_index('time')
            df['timestamp'] = df['corrected_timestamp'].dt.strftime('%m-%d %H:%M%')
            df['timestamp'] = pd.to_datetime(df['timestamp'])  
            #df['timestamp'] = df['timestamp'].dt.tz_localize(None)
            df = df.set_index('timestamp')
            
            dtStart = datetime.datetime(simulationDict['startHour'][0], simulationDict['startHour'][1], simulationDict['startHour'][2], simulationDict['startHour'][3], tzinfo=dateutil.tz.tzoffset(None, simulationDict['utcOffset']*60*60))
        
        
            dtEnd = datetime.datetime(simulationDict['endHour'][0], simulationDict['endHour'][1], simulationDict['endHour'][2], simulationDict['endHour'][3], tzinfo=dateutil.tz.tzoffset(None, simulationDict['utcOffset']*60*60))
            mask = (df.index >= dtStart) & (df.index <= dtEnd) 
            df = df.loc[mask]
        
        df['time'] = df['corrected_timestamp'].dt.strftime('%m_%d_%H')
        df = df.set_index('time')
        
        #Diode ideality factors. a1 has to be 1 while a2 is flexible but it has to be above 1.2
        a1 = 1      
        a2 = 1.3
        
        #Tolerances for current and power in Mpp calculation algorythm
        tol_I = 0.001 
        tol_P = 0.002
        
        #Calculation of Parameters that do not change in the loop
        P_mpp_f0 = V_mpp_f0 * I_mpp_f0                      #P_mpp is calculated from Mpp Voltage and Current
        Vt = (Ns * k * (25+273.15)) / q_ec                  #Thermal voltage at STC                                     
        I_0_f0 = (I_sc_f0) / (np.exp((V_oc_f0) / (Vt))-1)   #Calculation of saturation current. It is assumed that its the same for both diodes

        #Starting values fpr Rs and Rp
        Rs_f0 = 0
        Rp_min = (V_mpp_f0 / (I_sc_f0 - I_mpp_f0)) - ((V_oc_f0 - V_mpp_f0) / I_mpp_f0)
        Rp_f0 = Rp_min
        
        #Starting parameters for Current, Voltage, and Power
        I = 0
        V = 0
        P = 0
        P1 = 0
        P_mpp_fs0 = 0
        
        #Definition of lists for the I-V P-V plots
        Vfplt = []
        Ifplt = []
        Pfplt = []
               
        #Rs and Rp calculation for the front side. They are calculated at STC. The algorythm tries to match the Mpp of the Module parameters
        #by adjusting Rs and calculating Rp based on Rs.
        for xf in range (1000000):
           
            
           #If the calculation takes too many iterations, an error message will show up.
           if xf == 999999:                     
                print ('Error: The calculation of the front resistances exeeded one million iterations. Please check your parameters.')
                     
           #'Calculation' of the photo current. At STC the I_sc is the photo current.        
           I_ph_f0 = I_sc_f0                                             
       
           #Newtons method to find the current that solves the equation f_I  
           f_I = I_ph_f0 - I_0_f0 * (np.exp((V+ I * Rs_f0) / Vt) + np.exp((V + I * Rs_f0) / (Vt * a2)) - 2) - ((V + I * Rs_f0) / Rp_f0) - I
           f_dI = (-1) * ((I_0_f0 * Rs_f0) / Vt) * np.exp((V + I * Rs_f0) / Vt) - ((I_0_f0 * Rs_f0) / (Vt * a2)) * np.exp((V+ I * Rs_f0) / (Vt * a2)) - (Rs_f0 / Rp_f0) - 1
       
           I2 = I - (f_I/f_dI)
              
           #Check if the new value of I2 is close enough (in tolerance) to the old value I
           if I + tol_I >= I2 and I2 >= I - tol_I:
               P2 = V * I2          #If the current was calculated correctly the power will be calculated
               
               Vfplt.append(V)      #The values of current, voltage and power are added to their designated lists
               Ifplt.append(I2)
               Pfplt.append(P2)
               
               if P2 > P1:          #Check if the new power value is greater than the last. If yes it becomes the new reference power value.
                   P1 = P2
               
               else:
                   P_mpp_fs0 = P1   #If the power does not increase however, the Mpp of this P-V curve is found
                   
                   
                           
               if V >= V_oc_f0:     # When V reaches V_oc the P-V curve is complete and it has to be checked if Mpp converged with the Module parameters
               
                   #If the calculated Mpp is within the tolerance the script tells the user the Values of Rs and Rp aswell as the amount of iterations
                   #it took to calculate them                   
                   if P_mpp_f0 + tol_P >= P_mpp_fs0 and P_mpp_fs0 >= P_mpp_f0 - tol_P:
                       print ('Front resistance calculation completed in ',xf,' iterations.', 'Rs_f0 =', Rs_f0, 'Rp_f0 =', Rp_f0)
                       
                       
                       #The P-V and I-V curve is plotted with the values since the last Rs, Rp adjustment
                       f, (ax1) = plt.subplots(1, figsize=(6, 6))
                       ax1.locator_params(tight=True, nbins=6)
                       ax1.plot(Vfplt[-round(V_oc_f0 * 10):], Ifplt[-round(V_oc_f0 * 10):], color = "red")
                       ax1.set_title('P-V I-V Curve front side', fontsize=14)
                       ax1.set_xlabel('Voltage [V]', fontsize=14)
                       ax1.set_ylabel('Current [I]', fontsize=14, color = 'red')
                       ax2=ax1.twinx()
                       ax2.plot(Vfplt[-round(V_oc_f0 * 10):], Pfplt[-round(V_oc_f0 * 10):], color = "blue")
                       ax2.set_ylabel('Power [W]', fontsize=14, color = 'blue')
                       f.savefig("P-V_I-V_Curve_front" + datetime.datetime.now().strftime("%Y-%m-%d-%H-%M") + ".png", dpi = dpi)
                       #plt.show()()
 
                       #Rs and Rp have been calculated so the script quits the loop 
                       break
                   
                   #If the calculated Mpp is not within tolerance, the reference power is reset and Rs gets increased by 10 mOhm 
                   else:
                       P1 = 0
                   
                       Rs_f0 = Rs_f0 + 0.01
                   
                       #Calculation of Rp with new Rs. For a better overview it is split into smaller calculations
                       r1 = V_mpp_f0 + I_mpp_f0 * Rs_f0
                       r2 = I_ph_f0
                       r3 = np.exp(r1/Vt)
                       r4 = np.exp(r1/(Vt*a2))
                       r5 = P_mpp_f0/V_mpp_f0
                       r6 = (r3 + r4 + 2)
                       r7 = r6 * I_0_f0

                       Rp_f0 = r1 / (r2 - r7 - r5)
                                     
                   V = -0.1     #V reached its max so it is set to -0.1 so in the first iteration we have I=0 and V=0
               I = 0
               V = V + 0.1
           else:
               I = I2           #If I2 was not close enough to I, I becomes I2 (newtons method)
        

        
        sum_energy_m = 0
        sum_power_m = 0
        
        
        # Loop to calculate the Bifacial Output power for every row in every hour
        for i in tqdm(range(0, simulationDict['nRows'])):
            
            key_front = "row_" + str(i) + "_qabs_front"
            key_back = "row_" + str(i) + "_qabs_back"
            
            P_m_hourly = []
            P_bi_hourly = []
          
            for index, row in df_report.iterrows():
                
                row_qabs_front = df_report.loc[index,key_front]
                row_qabs_back = df_report.loc[index,key_back]
                T_Current = df.loc[index,'temperature']
                
                
                #print("front: " + str(row_qabs_front))
                #print("back: " + str(row_qabs_back))
                if math.isnan(row_qabs_front) or row_qabs_front < 0.0:
                    row_qabs_front = 0
                    
                if math.isnan(row_qabs_back) or row_qabs_back < 0.0:
                    row_qabs_back = 0

                
                if row_qabs_back + row_qabs_front > 0.0:
                    
                    #Now that Rs and Rp are calculated for both sides of the module, the power calculation starts.
                    #Values are now adjusted for temperature and later also irradiation
                    
                    I_sc_f = I_sc_f0 * (1 + T_koeff_I * (T_Current - T_amb))     
                    I_sc_r = I_sc_r0 * (1 + T_koeff_I * (T_Current - T_amb))
                    
                    V_oc_f = V_oc_f0 * (1 + T_koeff_V * (T_Current - T_amb))
                    V_oc_r = V_oc_r0 * (1 + T_koeff_V * (T_Current - T_amb))
                    
                    #setting starting parameters for the loop
                    I = 0
                    V = 0
                    P = 0
                    P1 = 0
                    P_mpp_sf = 0
                    
                    
                    #Modified version of the Rs, Rp calculation loop. Since Rs and Rp are now given, the loop needs only a fraction of iterations
                    #to calculate the power from a given irradiance and temperature
                    #The way this works is that the algorythm searches for the Mpp of the module with the given irradiance and temperature
                    #Just like a real PV system would do
                    #It 'draws' the P-V curve and finds the Mpp which is then the power output of the module
                    for yf in range (100000):
                        
                        if row_qabs_front == 0:
                            P_m = 0
                            P_mpp_sf = 0
                            break
                        
                        #Calculation of the photo current + correction for irrandiance and temperature
                        Vt_f = (Ns * k * (T_Current + 273.15)) / q_ec

                        #Calculation of the saturation current
                        I_0_f0 = (I_sc_f) / (np.exp((V_oc_f) / (Vt_f))-1)

                        #adjustment ot the photo current for irradiation
                        I_ph_f0 = I_sc_f                                             
                        I_ph_f = I_ph_f0 * (row_qabs_front / q_stc_front) 
              
                        #newthons method to find the matching current for a given voltage
                        f_I = I_ph_f - I_0_f0 * (np.exp((V+ I * Rs_f0) / Vt_f) + np.exp((V + I * Rs_f0) / (Vt_f * a2)) - 2) - ((V + I * Rs_f0) / Rp_f0) - I
                        f_dI = (-1) * ((I_0_f0 * Rs_f0) / Vt_f) * np.exp((V + I * Rs_f0) / Vt_f) - ((I_0_f0 * Rs_f0) / (Vt_f * a2)) * np.exp((V+ I * Rs_f0) / (Vt_f * a2)) - (Rs_f0 / Rp_f0) - 1
       
                        I2 = I - (f_I/f_dI)
              
       
                        if I + tol_I >= I2 and I2 >= I - tol_I:         #Once I is found, the Power check starts just like in xf loop
                            P2 = V * I2

                            if P2 > P1:                 #Check if the new power is higher than the last
                                P1 = P2                 #If this is true, it becomes the new reference value
               
                            else:
                                P_mpp_sf = P1           #The highest calculated power gets added to P_mpp_sf
                                
                            if V >= V_oc_f:             # If V reached V_oc, the P-V curve is complete and the Mpp can be searched
                                P_f = P_mpp_sf          #P_mpp_sf is the calculated Mpp for the Module for the given irradiance and and temperature
                                P_m = P_f
                                sum_energy_m += P_m     #The value gets added to a list for Bifacial gain calculation
                                yf = 0                  #Iterations get reset after successful Mpp calculation
                                break                   #The Mpp was found, the script quits the loop
                          
               
                            I = 0                       #Since there is only one P-V curve to calculate, V does not have to be reset
                            V = V + 0.1
                        else:
                            I = I2
                        
                            
                    ##################################
                    ###Same procedure for back side###
                    ##################################
                    I = 0
                    V = 0
                    P = 0
                    P1 = 0
                    
                    
                    
                    for yr in range (100000):
                        
                        if row_qabs_front == 0:
                            P_mpp_sr = 0
                            break
                  
                        #Calculation of the photo current + correction for irrandiance and temperature
                        Vt_f = (Ns * k * (T_Current + 273.15)) / q_ec
                        
                        #Calculating the bifacial gain factor
                        BG = row_qabs_back / row_qabs_front
                        
                       

                        
                        #Calculation of the saturation current
                        I_0_f0 = (I_sc_f) / (np.exp((V_oc_f) / (Vt_f))-1)

               

                        #adjustment ot the photo current for irradiation and bifacial gain factor
                        I_ph_f0 = I_sc_f *BG                                             
                        I_ph_f = I_ph_f0 * (row_qabs_front / q_stc_front) 
              
                        #newthons method to find the matching current for a given voltage
                        f_I = I_ph_f - I_0_f0 * (np.exp((V+ I * Rs_f0) / Vt_f) + np.exp((V + I * Rs_f0) / (Vt_f * a2)) - 2) - ((V + I * Rs_f0) / Rp_f0) - I
                        f_dI = (-1) * ((I_0_f0 * Rs_f0) / Vt_f) * np.exp((V + I * Rs_f0) / Vt_f) - ((I_0_f0 * Rs_f0) / (Vt_f * a2)) * np.exp((V+ I * Rs_f0) / (Vt_f * a2)) - (Rs_f0 / Rp_f0) - 1
       
                        I2 = I - (f_I/f_dI)
              
       
                        if I + tol_I >= I2 and I2 >= I - tol_I:         #Once I is found, the Power check starts just like in xf loop
                            P2 = V * I2

                            if P2 > P1:                 #Check if the new power is higher than the last
                                P1 = P2                 #If this is true, it becomes the new reference value
               
                            else:
                                P_mpp_sr = P1           #The highest calculated power gets added to P_mpp_sr
                                
                            if V >= V_oc_f:             # If V reached V_oc, the P-V curve is complete and the Mpp can be searched
                                P_r = P_mpp_sr          #P_mpp_sr is the calculated Mpp for the Module for the given irradiance and and temperature
                                yr = 0                  #Iterations get reset after successful Mpp calculation
                                break                   #The Mpp was found, the script quits the loop
                          
               
                            I = 0                       #Since there is only one P-V curve to calculate, V does not have to be reset
                            V = V + 0.1
                        else:
                            I = I2
                    
                
                    P_bi = P_mpp_sf + P_mpp_sr
                    #print("Power: " + str(P_bi))
            
                    sum_energy_b += P_bi # Sum up the energy of every row in every hour
            
                else:
                    P_m=0
                    P_bi=0
                
                P_m_hourly.append(P_m)
                P_bi_hourly.append(P_bi)
                
            # Append P_bi_hourly array to arrays
            P_m_hourly_arrays.append(P_m_hourly)
            P_bi_hourly_arrays.append(P_bi_hourly)
            
        P_bi_hourly_average = []
        
        for i in tqdm(range(0, len(P_bi_hourly_arrays[0]))):
            sum = 0
          
            for j in range(0, len(P_bi_hourly_arrays)):
                sum += P_bi_hourly_arrays[j][i]
                
            average = sum / float(len(P_bi_hourly_arrays))
            
            P_bi_hourly_average.append(average)
            
       
        
        # The time gets implemented in the GUI
    # p_bi_df.to_csv(resultsPath + "electrical_simulation" + datetime.datetime.now().strftime("%Y-%m-%d-%H-%M") + ".csv")
        
        
        annual_power_per_module_b = (sum_energy_b/simulationDict['nRows']) #[W] annual bifacial output power per module
        '''print("Yearly bifacial output power per module: " + str(annual_power_per_module_b) + " W/module")
        print("Yearly bifacial output energy per module: " + str(annual_power_per_module_b/1000) + " kWh/module") # Because the input data is per hour, the Energy is equivalent to the performance
        print ("\n")'''
        
        annual_power_per_peak_b = (sum_energy_b/(P_mpp0 * simulationDict['nRows']))   #[W/Wp] annual bifacial output power per module peak power
        '''print("Yearly bifacial output power per module peak power: " + str(annual_power_per_peak_b) + " W/Wp")
        print("Yearly bifacial output energy per module peak power: " + str(annual_power_per_peak_b) + " kWh/kWp") # Because the input data is per hour, the Energy is equivalent to the performance 
        print ("\n")'''
        
        module_area = (simulationDict['moduley'] *simulationDict['nModsy'] *simulationDict['modulex'])
        
        annual_power_per_area_b = (annual_power_per_module_b / module_area)    #[W/m^2] annual bifacial poutput power per module area
        '''print("Yearly bifacial output power per module area: " + str(annual_power_per_area_b) + " W/m^2")
        print("Yearly bifacial output energy per module area: " + str(annual_power_per_area_b/1000) + " kWh/m^2") # Because the input data is per hour, the Energy is equivalent to the performance 
        print ("\n")'''
        
        # Plot total qinc front and back for every row
      
        
        f = plt.Figure(figsize=(12, 3))
        ax1 = f.subplots(1)
        ax1.locator_params(tight=True, nbins=6)
        f.plot(P_bi_hourly)
        ax1.set_title('Bifacial output Power hourly')
        ax1.set_xlabel('Hour')
        ax1.set_ylabel('W')
        f.savefig("P_bi_hourly" + datetime.datetime.now().strftime("%Y-%m-%d-%H-%M") + ".png", dpi = dpi)
        #plt.show()()
        #sns wird entfernt da die Plots in der GUI ansonsten nicht richtig angezeigt werden können außerhalb der Konsole
        ##plt.show()(sns)
        
        
        ####################################################
        # Monofacial performance Calculation
        
        # Set Energy to Zero       
        #sum_energy_m = 0
        #sum_power_m = 0
        
        # Loop to calculate the Monofacial Output power for every row in every hour
        #for i in tqdm(range(0, simulationDict['nRows'])):
            
        #    key_front_mono = "row_" + str(i) + "_qabs_front"
        #    P_m_hourly = []
            
        #    for index, row in df_report.iterrows():
                
        #        #SG
        #        row_qabs_front = df_report.loc[index,key_front_mono]
        #        T_Current = df.loc[index,'temperature']

        #        if math.isnan(row_qabs_front):
        #            row_qabs_front = 0 
                
        #        if  row_qabs_front > 0.0:
              
        #            V_oc_f = V_oc_f0 * (1 + T_koeff_V * (T_Current - T_amb) + moduleDict['zeta'] * np.log(row_qabs_front / q_stc_front))
        #            I_sc_f = I_sc_f0 * (1 + T_koeff_I * (T_Current - T_amb)) * (row_qabs_front / q_stc_front)
        #            P_m = FF_f0 * V_oc_f * I_sc_f
                
                    #print("Power: " + str(P_bi))
             
        #            sum_energy_m += P_m # Sum up the energy of every row in every hour
        #        else:
        #            P_m = 0
                    
        #        P_m_hourly.append(P_m)
            
            # Append P_m_hourly array to arrays
        #    P_m_hourly_arrays.append(P_m_hourly)
        
        
        P_m_hourly_average = []
        
        for i in tqdm(range(0, len(P_m_hourly_arrays[0]))):
            sum = 0
          
            for j in range(0, len(P_m_hourly_arrays)):
                sum += P_m_hourly_arrays[j][i]
                
            average_m = sum / float(len(P_m_hourly_arrays))
            
            P_m_hourly_average.append(average_m)
                 #else:
                    #print("Power: 0.0")
        
        annual_power_per_module_m = (sum_energy_m/simulationDict['nRows']) #[W] annual monofacial output power per module
        '''print("Yearly monofacial output power per module: " + str(annual_power_per_module_m) + " W/module")
        print("Yearly monofacial output energy per module: " + str(annual_power_per_module_m/1000) + " kWh/module") # Because the input data is per hour, the Energy is equivalent to the performance
        print ("\n")'''
        
        annual_power_per_peak_m = (sum_energy_m/(P_mpp0 * simulationDict['nRows']))   #[W/Wp] annual monofacial output power per module peak power
        '''print("Yearly monofacial output power per module peak power: " + str(annual_power_per_peak_m) + " W/Wp")
        print("Yearly monofacial output energy per module peak power: " + str(annual_power_per_peak_m) + " kWh/kWp") # Because the input data is per hour, the Energy is equivalent to the performance 
        print ("\n")'''
        
        module_area = (simulationDict['moduley'] *simulationDict['nModsy']* simulationDict['modulex'])
        
        annual_power_per_area_m = (annual_power_per_module_m / module_area)    #[W/m^2] annual monofacial poutput power per module area
        '''print("Yearly monofacial output power per module area: " + str(annual_power_per_area_m) + " W/m^2")
        print("Yearly monofacial output energy per module area: " + str(annual_power_per_area_m/1000) + " kWh/m^2") # Because the input data is per hour, the Energy is equivalent to the performance 
        print ("\n")'''
        
        ####################################################
        
        # Bifacial Gain Calculation
        
        Bifacial_gain= (annual_power_per_peak_b - annual_power_per_peak_m) / annual_power_per_peak_m
        print("Bifacial Gain: " + str(Bifacial_gain*100) + " %")
        
                
        # Create dataframe with data
        p_bi_df = pd.DataFrame({"timestamps":df_report.index, "P_bi ": P_bi_hourly_average, "P_m ": P_m_hourly_average})
        p_bi_df.set_index("timestamps")
        p_bi_df.to_csv(resultsPath + "electrical_simulation" + datetime.datetime.now().strftime("%Y-%m-%d-%H-%M") + ".csv")
        #p_bi_df.to_csv(resultsPath + "electrical_simulation.csv")
        
        #Plot for Bifacial Power Output + Bifacial Gain
        GUI.Window.makePlotBifacialRadiance(resultsPath,Bifacial_gain)
        
        def simulate_simpleBifacial_old(moduleDict, simulationDict, df_reportVF, df_reportRT, df_report, df, resultsPath):
            """
            Applies a simplified version of the electrical simulation after PVSyst. Uses bifaciality factor to calculate rear efficiency and fill factors.
            Rear open-circuit voltage and short-circuit current are calculated using rear irradiance and temperature. 
            The rear output is then determined through the rear fill factor. Then the bifacial electrical output is calculated by adding front and rear output.
            Calculates bifacial gain through a seperate monofacial electrical simulation.

            Parameters
            ----------
            moduleDict: module Dictionary containing module data
            simulationDict: simulation Dictionary, which can be found in BifacialSimu_main.py
            df_reportVF: Viewfactor simulation report
            df_reportRT: Raytracing simulation report
            df_report: Final simulation report, containing VF and RT data
            resultsPath: output filepath
            df: helper DataFrame containing temperature for electrical simulation
            """
            
            
            # Build a final simulation report
            df_report = Electrical_simulation.build_simulationReport(df_reportVF, df_reportRT, simulationDict, resultsPath)
            
            ####################################################
            # Variables required for electrical simulation
            
            # P_bi: Output power of bifacial module for bifacial illumination (W)
            # I_sc_bi: Short-circuit current of bifacial module for bifacial illumination (A)
            # V_oc: Open-circuit voltage of bifacial module for bifacial illumination (V)
            # FF_bi: Fill factor of bifacial module for bifacial illumination (%)
            # G_r: Irradiance on the rear side of the module (W/m2)
            # G_f: Irradiance on the front side of the module (W/m2)
            # R_isc: Relative current gain (dimensionless)
            # I_sc_f: Short-circuit current measured for front side illumination of the module at STC (A)
            # x: Irradiance ratio (dimensionless)
            # V_oc_f: Open-circuit voltage measured for front side illumina- tion of module at STC (V)
            # V_oc_r: Open-circuit voltage measured for rear side illumina- tion of module at STC (V)
            # I_sc_r: Short-circuit current measured for rear side illumination of the module at STC (A)
            # FF_f: Fill factor measured for front side illumination of the module at STC (%)
            # FF_r: Fill factor measured for rear side illumination of the module (%)
            # pFF: Pseudo fill factor (FF of the module considering no series resistance effect) (%)

            ####################################################
            # Definition of simulation parameter if only front parameters are available
            # Procedure after PVSyst
            V_mpp_f0 = moduleDict['V_mpp_f']
                
            I_mpp_f0 = moduleDict['I_mpp_f']
                    
            I_sc_f0 = moduleDict['I_sc_f']
            
            V_oc_f0 = moduleDict['V_oc_f']
            
            #module = moduleParameter['module']
            #inverter = moduleParameter['inverter']
            
            P_mpp0 = moduleDict['P_mpp']
            V_mpp0 = V_mpp_f0
            
            T_koeff_P = moduleDict['T_koeff_P'] 
            T_koeff_I = moduleDict['T_koeff_I'] 
            T_koeff_V = moduleDict['T_koeff_V'] 
            T_amb = moduleDict['T_amb']
            
            q_stc_front = 1000  # [W/m^2] 
            q_stc_rear = 1000   # [W/m^2] 
            
            # Calculation of fill factor for STC conditions
            FF_f0 = (I_mpp_f0 * V_mpp_f0)/(I_sc_f0 * V_oc_f0)
            
            n_back = (moduleDict['bi_factor']*moduleDict['n_front'])
            
            FF_fr = (n_back*q_stc_rear*(simulationDict['moduley'])*(simulationDict['modulex']))/(I_sc_f0 * V_oc_f0)
                           
            dpi = 150 #Quality for plot export
            
            # Set Energy to Zero       
            sum_energy_b = 0
            sum_power_b = 0
            
            # Array to hold other arrays -> average after for loop
            P_bi_hourly_arrays = []
            
            df_report['timestamp'] = df_report.index
            df_report = df_report.reset_index()
            df_report['corrected_timestamp'] = pd.to_datetime(df_report['timestamp'])
            df_report['time'] = df_report['corrected_timestamp'].dt.strftime('%m_%d_%H')
            df_report = df_report.set_index('time')
            
            if simulationDict['simulationMode'] == 3:
                df = df.reset_index()
                
                df['time'] = df['corrected_timestamp'].dt.strftime('%m_%d_%H')
                df = df.set_index('time')
                df['timestamp'] = df['corrected_timestamp'].dt.strftime('%m-%d %H:%M%')
                df['timestamp'] = pd.to_datetime(df['timestamp'])  
                #df['timestamp'] = df['timestamp'].dt.tz_localize(None)
                df = df.set_index('timestamp')
                
                dtStart = datetime.datetime(simulationDict['startHour'][0], simulationDict['startHour'][1], simulationDict['startHour'][2], simulationDict['startHour'][3], tzinfo=dateutil.tz.tzoffset(None, simulationDict['utcOffset']*60*60))
            
            
                dtEnd = datetime.datetime(simulationDict['endHour'][0], simulationDict['endHour'][1], simulationDict['endHour'][2], simulationDict['endHour'][3], tzinfo=dateutil.tz.tzoffset(None, simulationDict['utcOffset']*60*60))
                mask = (df.index >= dtStart) & (df.index <= dtEnd) 
                df = df.loc[mask]
            
            df['time'] = df['corrected_timestamp'].dt.strftime('%m_%d_%H')
            df = df.set_index('time')
            print(df_report)
            
            
            
            # Loop to calculate the Monofacial Output power for every row in every hour
            for i in tqdm(range(0, simulationDict['nRows'])):
                
                key_front = "row_" + str(i) + "_qabs_front"
                key_back = "row_" + str(i) + "_qabs_back"
            
                P_bi_hourly = []
              
                for index, row in df_report.iterrows():
                    
                    #row_qabs_front = row[key_front]
                    #row_qabs_back = row[key_back]
                    
                    row_qabs_front = df_report.loc[index,key_front]
                    row_qabs_back = df_report.loc[index,key_back]
                    T_Current = df.loc[index,'temperature']
                    
                    
                    # calculation of frontside power output
                    if math.isnan(row_qabs_front) or row_qabs_front < 0.0:
                        row_qabs_front = 0
                        P_f = 0
                        
                    else:
                        V_oc_f = V_oc_f0 * (1 + T_koeff_V * (T_Current - T_amb) + moduleDict['zeta'] * np.log(row_qabs_front / q_stc_front))
                        I_sc_f = I_sc_f0 * (1 + T_koeff_I * (T_Current - T_amb)) * (row_qabs_front / q_stc_front)
                        P_f = FF_f0 * V_oc_f * I_sc_f

                    # calculation of backside power output
                    if math.isnan(row_qabs_back) or row_qabs_back < 0.0:
                        row_qabs_back = 0
                        P_r = 0
                        
                    else:
                        V_oc_r = V_oc_f0 * (1 + T_koeff_V * (T_Current - T_amb) + moduleDict['zeta'] * np.log(row_qabs_back / q_stc_rear))
                        I_sc_r = I_sc_f0 * (1 + T_koeff_I * (T_Current - T_amb)) * (row_qabs_back / q_stc_rear)
                        P_r = FF_fr * V_oc_r * I_sc_r
                        
                    
                    P_bi = P_f + P_r 
                    
            
                    
                    sum_energy_b += P_bi # Sum up the energy of every row in every hour

                    P_bi_hourly.append(P_bi)
                    
                # Append P_bi_hourly array to arrays
                P_bi_hourly_arrays.append(P_bi_hourly)

                print(sum_energy_b)
                    
            P_bi_hourly_average = []
            
            for i in tqdm(range(0, len(P_bi_hourly_arrays[0]))):
                sum = 0
              
                for j in range(0, len(P_bi_hourly_arrays)):
                    sum += P_bi_hourly_arrays[j][i]
                    
                average = sum / float(len(P_bi_hourly_arrays))
                
                P_bi_hourly_average.append(average)
                
                    
            # Create dataframe with average data
            p_bi_df = pd.DataFrame({"timestamps":df_report.index, "P_bi ": P_bi_hourly_average})
            p_bi_df.set_index("timestamps")
            p_bi_df.to_csv(resultsPath + "electrical_simulation" + datetime.datetime.now().strftime("%Y-%m-%d-%H-%M") + ".csv")
            
            
            annual_power_per_module_b = (sum_energy_b/simulationDict['nRows']) #[W] annual bifacial output power per module
            '''print("Yearly bifacial output power per module: " + str(annual_power_per_module_b) + " W/module")
            print("Yearly bifacial output energy per module: " + str(annual_power_per_module_b/1000) + " kWh/module") # Because the input data is per hour, the Energy is equivalent to the performance
            print ("\n")'''
            
            annual_power_per_peak_b = (sum_energy_b/(P_mpp0 * simulationDict['nRows']))   #[W/Wp] annual bifacial output power per module peak power
            '''print("Yearly bifacial output power per module peak power: " + str(annual_power_per_peak_b) + " W/Wp")
            print("Yearly bifacial output energy per module peak power: " + str(annual_power_per_peak_b) + " kWh/kWp") # Because the input data is per hour, the Energy is equivalent to the performance 
            print ("\n")'''
            
            module_area = (simulationDict['moduley'] * simulationDict['nModsy'] * simulationDict['modulex'])
            
            annual_power_per_area_b = (annual_power_per_module_b / module_area)    #[W/m^2] annual bifacial poutput power per module area
            '''print("Yearly bifacial output power per module area: " + str(annual_power_per_area_b) + " W/m^2")
            print("Yearly bifacial output energy per module area: " + str(annual_power_per_area_b/1000) + " kWh/m^2") # Because the input data is per hour, the Energy is equivalent to the performance 
            print ("\n")'''
            
            # Plot total qinc front and back for every row
          
            
            f = plt.Figure(figsize=(12, 3))
            ax1 = f.subplots(1)
            ax1.locator_params(tight=True, nbins=6)
            f.plot(P_bi_hourly)
            ax1.set_title('Bifacial output Power hourly')
            ax1.set_xlabel('Hour')
            ax1.set_ylabel('W')
            f.savefig("P_bi_hourly" + datetime.datetime.now().strftime("%Y-%m-%d-%H-%M") + ".png", dpi = dpi)
            #plt.show()()
            #sns wird entfernt da die Plots in der GUI ansonsten nicht richtig angezeigt werden können außerhalb der Konsole
            ##plt.show()(sns)
             
            ####################################################
            # Monofacial performance Calculation
            
            # Set Energy to Zero       
            sum_energy_m = 0
            sum_power_m = 0
            
            # Loop to calculate the Monofacial Output power for every row in every hour
            for i in tqdm(range(0, simulationDict['nRows'])):
                
                key_front_mono = "row_" + str(i) + "_qabs_front"
                
                for index, row in df_report.iterrows():
                    
                    #SG
                    #row_qabs_front = row[key_front_mono]
                    #T_Current = df.loc[index,'temperature']
                    row_qabs_front = df_report.loc[index,key_front_mono]
                    T_Current = df.loc[index,'temperature']

                    if math.isnan(row_qabs_front):
                        row_qabs_front = 0     
                    
                    if  row_qabs_front > 0.0:
                  
                        V_oc_f = V_oc_f0 * (1 + T_koeff_V * (T_Current - T_amb) + moduleDict['zeta'] * np.log(row_qabs_front / q_stc_front))
                        I_sc_f = I_sc_f0 * (1 + T_koeff_I * (T_Current - T_amb)) * (row_qabs_front / q_stc_front)
                        P_m = FF_f0 * V_oc_f * I_sc_f
                    
                 
                        sum_energy_m += P_m # Sum up the energy of every row in every hour
                    

            
            annual_power_per_module_m = (sum_energy_m/simulationDict['nRows']) #[W] annual monofacial output power per module
            '''print("Yearly monofacial output power per module: " + str(annual_power_per_module_m) + " W/module")
            print("Yearly monofacial output energy per module: " + str(annual_power_per_module_m/1000) + " kWh/module") # Because the input data is per hour, the Energy is equivalent to the performance
            print ("\n")'''
            
            annual_power_per_peak_m = (sum_energy_m/(P_mpp0 * simulationDict['nRows']))   #[W/Wp] annual monofacial output power per module peak power
            '''print("Yearly monofacial output power per module peak power: " + str(annual_power_per_peak_m) + " W/Wp")
            print("Yearly monofacial output energy per module peak power: " + str(annual_power_per_peak_m) + " kWh/kWp") # Because the input data is per hour, the Energy is equivalent to the performance 
            print ("\n")'''
            
            module_area = (simulationDict['moduley'] *simulationDict['nModsy']* simulationDict['modulex'])
            
            annual_power_per_area_m = (annual_power_per_module_m / module_area)    #[W/m^2] annual monofacial poutput power per module area
            '''print("Yearly monofacial output power per module area: " + str(annual_power_per_area_m) + " W/m^2")
            print("Yearly monofacial output energy per module area: " + str(annual_power_per_area_m/1000) + " kWh/m^2") # Because the input data is per hour, the Energy is equivalent to the performance 
            print ("\n")'''
            
            ####################################################
            
            # Bifacial Gain Calculation
            
            Bifacial_gain= (annual_power_per_peak_b - annual_power_per_peak_m) / annual_power_per_peak_m
            print("Bifacial Gain: " + str(Bifacial_gain*100) + " %")
            
            #Plot for Bifacial Power Output + Bifacial Gain
            GUI.Window.makePlotBifacialRadiance(resultsPath,Bifacial_gain)         
            return Bifacial_gain*100


# -*- coding: utf-8 -*-
"""
Created on Mon Jun  7 11:39:16 2021
@author:        
    Eva-Maria Grommes

Additional co-authors can be found here:
https://github.com/cire-thk/bifacialSimu    

name:
    BifacialSimu - SimulationController

overview:
    Read the input data and settings from the main script, calls the functions 
    of the handlers to perform the bifacial simulation of PV-Modules with
    View Factors and/or Ray Tracing method


"""

'''import os
import sys
# Path handling
rootPath = os.path.dirname(os.path.dirname(os.path.realpath(".")))
print(rootPath)
# Include paths
sys.path.append(rootPath)
sys.path.append(rootPath + "/BifacialSimu/Handler")'''

import pandas as pd
import sys
from BifacialSimu.Handler import * #much easier handling Directories using __init__.py files (avoids import errors)

# Overarching procedure to perform bifacial irrdiance and electrical simulations  
def startSimulation(simulationDict, moduleDict, resultsPath):

    #the path is implemented in GUI.py
    # resultsPath = BifacialSimu_dataHandler.DataHandler().setDirectories()
    # print('created resultsPath at: ' + resultsPath)

    
    #get weatherFile
    metdata, demo = BifacialSimu_dataHandler.DataHandler().getWeatherData(simulationDict, resultsPath)

    print('succsessfully created metdata and demo')
    
    # pass weatherfile to df
    df = BifacialSimu_dataHandler.DataHandler().passEPWtoDF(metdata, simulationDict, resultsPath)
    df_reportRT = pd.DataFrame()
    df_reportVF = pd.DataFrame()
    df_report = pd.DataFrame()
    
    ####################################################
    
    # optional spectralAlbedo calculation
    if simulationDict['hourlySpectralAlbedo'] == True:
        # spectralAlbedoHandler calculate the spectral albedo and write it in the weatherfile in colume 'albedo'
        #BifacialSimu_spectralAlbedoHandler_1_row.calculateAlbedo(simulationDict, df, resultsPath)
        BifacialSimu_spectralAlbedoHandler.calculateAlbedo(simulationDict, df, resultsPath)
        # weatherfile is read in again with updated albedo values as metdata
        metdata, demo = BifacialSimu_dataHandler.DataHandler().getWeatherData(simulationDict, resultsPath) 
    
        # metdata is converted to df
        df = BifacialSimu_dataHandler.DataHandler().passEPWtoDF(metdata, simulationDict, resultsPath)
        print('succsessfully updated metdata, demo and df with spectral albedo')
    
    ####################################################
    
    # choose simulation mode and perform raytracing, viewfactor and electrical simulation
    
    if simulationDict['simulationMode'] == 3:
        print('Front and back simulation with RayTrace')
        df_reportRT = BifacialSimu_radiationHandler.RayTrace.simulateRayTrace(simulationDict, demo, metdata, resultsPath, df, onlyBackscan = False)
        
        if simulationDict['ElectricalMode_simple'] == 0:      
            BifacialSimu_calculationHandler.Electrical_simulation.simulate_simpleBifacial(moduleDict, simulationDict, df_reportVF, df_reportRT, df_report, df, resultsPath)
        if simulationDict['ElectricalMode_simple'] == 1:
            BifacialSimu_calculationHandler.Electrical_simulation.simulate_oneDiode(moduleDict, simulationDict, df_reportVF, df_reportRT, df_report, df, resultsPath)
        if simulationDict['ElectricalMode_simple'] == 2:      
            BifacialSimu_calculationHandler.Electrical_simulation.simulate_doubleDiode(moduleDict, simulationDict, df_reportVF, df_reportRT, df_report, df, resultsPath)
        if simulationDict['ElectricalMode_simple'] == 3:
            BifacialSimu_calculationHandler.Electrical_simulation.simulate_doubleDiodeBi(moduleDict, simulationDict, df_reportVF, df_reportRT, df_report, df, resultsPath)
        
    if simulationDict['simulationMode'] == 5 or simulationDict['simulationMode'] == 1:
        print('Back simulation with RayTrace')
        df_reportRT =  BifacialSimu_radiationHandler.RayTrace.simulateRayTrace(simulationDict, demo, metdata, resultsPath, df, onlyBackscan = True)
        
        
    if simulationDict['simulationMode'] == 2:
        print('Front and back simulation with ViewFactors')
        df_reportVF, df = BifacialSimu_radiationHandler.ViewFactors.simulateViewFactors(simulationDict, demo, metdata,  df, resultsPath, onlyFrontscan = False)
        
        if simulationDict['ElectricalMode_simple'] == 0:      
            BifacialSimu_calculationHandler.Electrical_simulation.simulate_simpleBifacial(moduleDict, simulationDict, df_reportVF, df_reportRT, df_report, df, resultsPath)
        if simulationDict['ElectricalMode_simple'] == 1:
            BifacialSimu_calculationHandler.Electrical_simulation.simulate_oneDiode(moduleDict, simulationDict, df_reportVF, df_reportRT, df_report, df, resultsPath)
        if simulationDict['ElectricalMode_simple'] == 2:      
            BifacialSimu_calculationHandler.Electrical_simulation.simulate_doubleDiode(moduleDict, simulationDict, df_reportVF, df_reportRT, df_report, df, resultsPath)
        if simulationDict['ElectricalMode_simple'] == 3:
            BifacialSimu_calculationHandler.Electrical_simulation.simulate_doubleDiodeBi(moduleDict, simulationDict, df_reportVF, df_reportRT, df_report, df, resultsPath)
    
    
    if simulationDict['simulationMode'] == 4 or simulationDict['simulationMode'] == 1:
        print('Front simulation with ViewFactors')
        df_reportVF, df = BifacialSimu_radiationHandler.ViewFactors.simulateViewFactors(simulationDict, demo, metdata,  df, resultsPath, onlyFrontscan = True)
        
        if simulationDict['ElectricalMode_simple'] == 0:      
            BifacialSimu_calculationHandler.Electrical_simulation.simulate_simpleBifacial(moduleDict, simulationDict, df_reportVF, df_reportRT, df_report, df, resultsPath)
        if simulationDict['ElectricalMode_simple'] == 1:
            BifacialSimu_calculationHandler.Electrical_simulation.simulate_oneDiode(moduleDict, simulationDict, df_reportVF, df_reportRT, df_report, df, resultsPath)
        if simulationDict['ElectricalMode_simple'] == 2:      
            BifacialSimu_calculationHandler.Electrical_simulation.simulate_doubleDiode(moduleDict, simulationDict, df_reportVF, df_reportRT, df_report, df, resultsPath)
        if simulationDict['ElectricalMode_simple'] == 3:
            BifacialSimu_calculationHandler.Electrical_simulation.simulate_doubleDiodeBi(moduleDict, simulationDict, df_reportVF, df_reportRT, df_report, df, resultsPath)



# import sys
import os
rootPath = os.path.realpath("../")

import unittest
import pandas as pd
from BifacialSimu.Handler.BifacialSimu_calculationHandler import Electrical_simulation
from BifacialSimu.Handler.BifacialSimu_radiationHandler import ViewFactors
from BifacialSimu.Handler.BifacialSimu_dataHandler import DataHandler


# P.S. Values of simulationDict were made sure to be equal to the values of the default simulation Variables!
# fixing the values of simulationDict exactly to values of default Simulation variable, removes the AssertionError at level of BifacialSimu_radiationHandler!
simulationDict={
    'clearance_height': 0.4, #value was found missing! should be added later!
    'simulationName' : 'NREL_best_field_row_2',
    'simulationMode' : 1, 
    'localFile' : True, # Decide wether you want to use a  weather file or try to download one for the coordinates
    'weatherFile' : (rootPath +'/WeatherData/Golden_USA/SRRLWeatherdata Nov_Dez_2.csv'), #weather file in TMY format 
    'spectralReflectancefile' : (rootPath + '/ReflectivityData/interpolated_reflectivity.csv'),
    'cumulativeSky' : False, # Mode for RayTracing: CumulativeSky or hourly
    'startHour' : (2019, 11, 1, 0),  # Only for hourly simulation, yy, mm, dd, hh
    'endHour' : (2019, 11, 16, 0),  # Only for hourly simulation, yy, mm, dd, hh
    'utcOffset': -7,
    'tilt' : 25, #tilt of the PV surface [deg]
    'singleAxisTracking' : False, # singleAxisTracking or not
    'backTracking' : False, # Solar backtracking is a tracking control program that aims to minimize PV panel-on-panel shading 
    'ElectricalMode_simple': 1, # simple electrical Simulation after PVSyst, use if rear module parameters are missing
    'limitAngle' : 60, # limit Angle for singleAxisTracking
    'hub_height' : 1.3, # Height of the rotation axis of the tracker [m]
    'azimuth' : 180, #azimuth of the PV surface [deg] 90°: East, 135° : South-East, 180°:South
    'nModsx' : 5, #number of modules in x-axis
    'nModsy' : 1, #number of modules in y-axis
    'nRows' : 3, #number of rows
    'sensorsy' : 5, #number of sensors
    'moduley' : 1.98 ,#length of modules in y-axis
    'modulex' : 0.992, #length of modules in x-axis  
    'fixAlbedo': True, # Option to use the fix albedo
    'hourlyMeasuredAlbedo' : True, # True if measured albedo values in weather file
    'hourlySpectralAlbedo' : True, # Option to calculate a spectral Albedo 
    'variableAlbedo': False, # Option to calculate sun position dependend, variable albedo
    'albedo' : 0.247, # Measured Albedo average value, if hourly isn't available
    'frontReflect' : 0.03, #front surface reflectivity of PV rows
    'BackReflect' : 0.05, #back surface reflectivity of PV rows
    'longitude' : -105.172, 
    'latitude' : 39.739,
    'gcr' : 0.45, #ground coverage ratio (module area / land use)
    'module_type' : 'NREL row 2', #Name of Module
    }

simulationName = simulationDict['simulationName']
onlyFrontscan=False
onlyBackscan=True
resultsPath = DataHandler().setDirectories()
 
metdata, demo = DataHandler().getWeatherData(simulationDict, resultsPath)

df = DataHandler().passEPWtoDF(metdata, simulationDict, resultsPath)
df_reportRT = pd.DataFrame()
df_reportVF = pd.DataFrame()
df_report = pd.DataFrame()
dataFrame = pd.DataFrame()

# df_report = Electrical_simulation.build_simulationReport(df_reportVF, df_reportRT, simulationDict, resultsPath)

# df_reportVF=ViewFactors.simulateViewFactors(simulationDict, demo, metdata, dataFrame, resultsPath, onlyFrontscan)
# df_reportRT = RayTrace.simulateRayTrace(simulationDict,demo, metdata, resultsPath, dataFrame, onlyBackscan)

moduleDict={
		"bi_factor": 0.694,
		"n_front": 0.19,
        "I_sc_f": 9.5,
        "I_sc_r": 6.56,
        "V_oc_f": 48,
        "V_oc_r": 47.3,
        "V_mpp_f": 39.2,
        "V_mpp_r": 39.5,
        "I_mpp_f": 9.0,
        "I_mpp_r": 6.2,
        "P_mpp": 354,
        "T_koeff_P": -0.0036,
        "T_amb": 20,
        "T_koeff_I": 0.0005,
        "T_koeff_V": 0.0005,
        "zeta": 0.06,
		"moduley": 1.98,
		"modulex": 0.992
}
   
    

class TestElectricalModes(unittest.TestCase):
    
    def test_simulate_simpleBifacial(self):
        df = DataHandler().passEPWtoDF(metdata, simulationDict, resultsPath)
        simulationDict['simulationMode'] = 2
        simulationDict['ElectricalMode_simple'] = 0
        df_reportVF, df= ViewFactors.simulateViewFactors(simulationDict, demo, metdata,  df, resultsPath, onlyFrontscan)
        test_result= Electrical_simulation.simulate_simpleBifacial(moduleDict, simulationDict, df_reportVF, df_reportRT, df_report, df, resultsPath)

        self.assertEqual(test_result, 5.599072940263201)

    def test_oneDiode(self):
        df = DataHandler().passEPWtoDF(metdata, simulationDict, resultsPath)
        simulationDict['simulationMode'] = 2
        simulationDict['ElectricalMode_simple'] = 1
        df_reportVF, df = ViewFactors.simulateViewFactors(simulationDict, demo, metdata,  df, resultsPath, onlyFrontscan = False)
        test_result= Electrical_simulation.simulate_oneDiode(moduleDict, simulationDict, df_reportVF, df_reportRT, df_report, df, resultsPath)
        
        self.assertEqual(test_result, 9.457514227805062)
    
    ''' simulate_doubleDiode() is still not implimented in GUI.py'''
    # def test_simulate_doubleDiode(self):
    #     df = DataHandler().passEPWtoDF(metdata, simulationDict, resultsPath)
    #     simulationDict['simulationMode'] = 2
    #     simulationDict['ElectricalMode_simple'] = 2
    #     df_reportVF, df = ViewFactors.simulateViewFactors(simulationDict, demo, metdata,  df, resultsPath, onlyFrontscan = False)
    #     test_result= Electrical_simulation.simulate_doubleDiode(moduleDict, simulationDict, df_reportVF, df_reportRT, df_report, df, resultsPath)
    #     self        

if __name__ == '__main__':
    unittest.main()


import numpy as np
import pandas as pd
from collections.abc import Iterable
import os
from scipy import integrate


class spectral_property(object):
    """
    WRITE DOCSTRING HERE
    """
    
    def load_file(filepath):
        with open(filepath, 'r') as infile:
            meta = next(infile)[:-1]
            data = pd.read_csv(infile)
        
        return spectral_property(data['value'], data['wavelength'],
                                 interpolation=meta.split(':')[1])
    
    def to_nm(wavelength, units):
        unit_conversion = { 'nm': 1,
                            'um': 1000 }
        
        # Verify units are in conversion table
        if units not in unit_conversion:
            print("Warning: Unknown unit specified. Options are {}.".format(
                unit_conversion.keys()))
            units = 'nm'
            
        return wavelength * unit_conversion[units]
    
    def _linear_interpolation(self, wavelength_nm):
        # Find upper and lower index
        upper_bound = self.data[self.data.index > wavelength_nm].index.min()
        lower_bound = self.data[self.data.index < wavelength_nm].index.max()
        
        # Determine values of surrounding indices
        upper_val = self.data['value'][upper_bound]
        lower_val = self.data['value'][lower_bound]
        
        # Calculate deltas
        delta_lambda = upper_bound - lower_bound
        delta_val = upper_val - lower_val
        
        return lower_val + delta_val*(wavelength_nm - lower_bound)/delta_lambda
    
    def _nearest_interpolation(self, wavelength_nm):
        # Find upper and lower index
        upper_bound = self.data[self.data.index > wavelength_nm].index.min()
        lower_bound = self.data[self.data.index < wavelength_nm].index.max()
        
        # Determine which index is closer
        if (upper_bound - wavelength_nm) < (wavelength_nm - lower_bound):
            return self.data['value'][upper_bound]
        return self.data['value'][lower_bound]
    
    def _lower_interpolation(self, wavelength_nm):
        # Find lower index
        lower_bound = self.data[self.data.index < wavelength_nm].index.max()
        
        return self.data['value'][lower_bound]
    
    def _upper_interpolation(self, wavelength_nm):
        # Find upper index
        upper_bound = self.data[self.data.index > wavelength_nm].index.min()
        
        return self.data['value'][upper_bound]
    
    interpolation_methods = {
        'linear':   _linear_interpolation,
        'nearest':  _nearest_interpolation,
        'lower':    _lower_interpolation,
        'upper':    _upper_interpolation
    }
    
    def __init__(self, values, index, index_units='nm', interpolation=None):
        # Verify lengths match
        if len(values) != len(index):
            print("Warning: Length of values and index must match.")
            return
        
        # Convert inputs to list
        values = [ val for val in values ]
        index = [ spectral_property.to_nm(idx, index_units) for idx in index ]
        
        # Create DataFrame
        self.data = pd.DataFrame()
        self.data['value'] = values
        self.data['wavelength'] = index
        self.data = self.data.set_index('wavelength')
        
        self.interpolation = None
        if interpolation in spectral_property.interpolation_methods:
            self.interpolation = \
                    spectral_property.interpolation_methods[interpolation]
            self.interpolation_type = interpolation
        elif interpolation:
            print("Warning: Specified interpolation type unknown.")
            
    def _get_single(self, wavelength, units):
        # Convert wavelength to nm
        wavelength = spectral_property.to_nm(wavelength, units)
        
        if wavelength in self.data.index:
            # If the value for that wavelength is known, return it
            return self.data['value'][wavelength]
        elif self.interpolation:
            # Check wavelength is within range
            if wavelength < self.data.index.min() or \
               wavelength > self.data.index.max():
                print("Warning: Requested wavelength outside spectrum.")
                return None
            
            # Return interpolated value
            return self.interpolation(self, wavelength)
        
        return None
    
    def __getitem__(self, wavelength, units='nm'):
        if isinstance(wavelength, Iterable):
            return np.array([ self._get_single(wl, units) for wl in wavelength ])
        return self._get_single(wavelength, units)
    
    def to_file(self, filepath, append=False):
        mode = 'w'
        if append:
            mode = 'a'
            
        with open(filepath, mode) as outfile:
            outfile.write(f"interpolation:{self.interpolation_type}\n")
            self.data.to_csv(outfile)
    
    def range(self):
        # Find upper and lower index
        upper_bound = self.data.index.max()
        lower_bound = self.data.index.min()
        
        return (lower_bound, upper_bound)
    
    def scale_values(self, scaling_factor):
        self.data['value'] *= scaling_factor
    
def spectral_albedo_smarts(zen, azm, material, min_wavelength=300,
                           max_wavelength=4000):

    import pySMARTS
    
    smarts_res = pySMARTS.SMARTSSpectraZenAzm('30 31', str(zen), str(azm), material,
                                     min_wvl=str(min_wavelength),
                                     max_wvl=str(max_wavelength))
    
    return spectral_property(smarts_res['Zonal_ground_reflectance'],
                             smarts_res['Wvlgth'], interpolation='linear')

def spectral_irradiance_smarts(zen, azm, material='LiteSoil', min_wavelength=300,
                           max_wavelength=4000):

    import pySMARTS

    smarts_res = pySMARTS.SMARTSSpectraZenAzm('2 3 4', str(zen), str(azm),
                                     material=material,
                                     min_wvl=str(min_wavelength),
                                     max_wvl=str(max_wavelength))
    
    dni_spectrum = spectral_property(smarts_res['Direct_normal_irradiance'],
                                     smarts_res['Wvlgth'], interpolation='linear')
    dhi_spectrum = spectral_property(smarts_res['Difuse_horizn_irradiance'],
                                     smarts_res['Wvlgth'], interpolation='linear')
    ghi_spectrum = spectral_property(smarts_res['Global_horizn_irradiance'],
                                     smarts_res['Wvlgth'], interpolation='linear')
    
    return (dni_spectrum, dhi_spectrum, ghi_spectrum)



def spectral_irradiance_smarts_SRRL(YEAR, MONTH, DAY, HOUR, ZONE,
                                LATIT, LONGIT, ALTIT,
                                RH, TAIR, SEASON, TDAY, SPR, W,
                                TILT, WAZIM, HEIGHT,
                                ALPHA1, ALPHA2, OMEGL, GG, BETA, TAU5,
                                RHOG, material,
                                IOUT='2 3 4', min_wvl='280', max_wvl='4000'):
    
    import pySMARTS

    smarts_res = pySMARTS.SMARTSSRRL(IOUT=IOUT, YEAR=YEAR,MONTH=MONTH,DAY=DAY,HOUR=HOUR, ZONE=ZONE,
                            LATIT=LATIT, LONGIT=LONGIT, ALTIT=ALTIT, 
                             RH=RH, TAIR=TAIR, SEASON=SEASON, TDAY=TDAY, SPR=SPR, W=W, 
                             TILT=TILT, WAZIM=WAZIM, HEIGHT=HEIGHT,
                             ALPHA1 = ALPHA1, ALPHA2 = ALPHA2, OMEGL = OMEGL,
                             GG = GG, BETA = BETA, TAU5= TAU5, 
                             RHOG=RHOG, material=material, 
                             min_wvl=min_wvl, max_wvl=max_wvl)
    

    dni_spectrum = spectral_property(smarts_res[smarts_res.keys()[1]],
                                     smarts_res['Wvlgth'], interpolation='linear')
    dhi_spectrum = spectral_property(smarts_res[smarts_res.keys()[2]],
                                     smarts_res['Wvlgth'], interpolation='linear')
    ghi_spectrum = spectral_property(smarts_res[smarts_res.keys()[3]],
                                     smarts_res['Wvlgth'], interpolation='linear')
    
    return (dni_spectrum, dhi_spectrum, ghi_spectrum)



def spectral_albedo_smarts_SRRL(YEAR, MONTH, DAY, HOUR, ZONE,
                                LATIT, LONGIT, ALTIT,
                                RH, TAIR, SEASON, TDAY, SPR, W,
                                TILT, WAZIM, HEIGHT, 
                                ALPHA1, ALPHA2, OMEGL, GG, BETA, TAU5,
                                RHOG, material,
                                IOUT='30 31', min_wvl='280', max_wvl='4000'):
 
    import pySMARTS

    smarts_res = pySMARTS.SMARTSSRRL(IOUT=IOUT, YEAR=YEAR,MONTH=MONTH,DAY=DAY,HOUR=HOUR, ZONE=ZONE,
                            LATIT=LATIT, LONGIT=LONGIT, ALTIT=ALTIT, 
                             RH=RH, TAIR=TAIR, SEASON=SEASON, TDAY=TDAY, SPR=SPR, W=W, 
                             TILT=TILT, WAZIM=WAZIM, HEIGHT=HEIGHT,
                             ALPHA1 = ALPHA1, ALPHA2 = ALPHA2, OMEGL = OMEGL,
                             GG = GG, BETA = BETA, TAU5= TAU5,
                             RHOG=RHOG, material=material, 
                             min_wvl=min_wvl, max_wvl=max_wvl)
    
    return spectral_property(smarts_res['Zonal_ground_reflectance'],
                             smarts_res['Wvlgth'], interpolation='linear')
    

def generate_spectra(idx, metdata, material=None, spectra_folder=None, scale_spectra=False,
                scale_albedo=False, scale_albedo_nonspectral_sim=False):
    """
    generate spectral curve for particular material.  Requires pySMARTS 

    Parameters
    ----------
    idx : int
        index of the metdata file to run pySMARTS.
    metdata : bifacial_radiance MetObj
        DESCRIPTION.
    material : string, optional
        type of material for spectral simulation. Options include: Grass, 
        Gravel, etc.     The default is None.
    spectra_folder : path, optional
        location to save spectral data. The default is None.
    scale_spectra : bool, optional
        DESCRIPTION. The default is False.
    scale_albedo : bool, optional
        DESCRIPTION. The default is False.
    scale_albedo_nonspectral_sim : bool, optional
        DESCRIPTION. The default is False.

    Returns
    -------
    spectral_alb : spectral_property class
        spectral_alb.data:  dataframe with frequency and magnitude data.
    spectral_dni : spectral_property class
        spectral_dni.data:  dataframe with frequency and magnitude data.
    spectral_dhi : spectral_property class
        spectral_dhi.data:  dataframe with frequency and magnitude data.

    """
    
    if material is None:
        material = 'Gravel'
    
    # Extract data from metdata
    dni = metdata.dni[idx]
    dhi = metdata.dhi[idx]
    ghi = metdata.ghi[idx]
    try:
        alb = metdata.albedo[idx]
    except TypeError:
        raise Exception("Error - No 'metdata.albedo' value passed.")

    solpos = metdata.solpos.iloc[idx]
    zen = float(solpos.zenith)
    azm = float(solpos.azimuth) - 180
    #lat = metdata.latitude
    #lon = metdata.longitude
    #elev = metdata.elevation / 1000
    #t = metdata.datetime[idx]
    
    # Verify sun up
    if zen > 90:
        print("Sun below horizon. Skipping.")
        return None
    
    # Define file suffix
    # -- CHANGE --
    suffix = f'_{idx:04}.txt'
    
    # Generate/Load dni and dhi
    dni_file = os.path.join(spectra_folder, "dni"+suffix)
    dhi_file = os.path.join(spectra_folder, "dhi"+suffix)
    ghi_file = os.path.join(spectra_folder, "ghi"+suffix)
    spectral_dni, spectral_dhi, spectral_ghi = spectral_irradiance_smarts(zen, azm, min_wavelength=300)
    
    # SCALING:
    # If specifed, scale the irradiance spectra based on their respective
    # measured value.
    if scale_spectra:
        
        dni_scale = dni / spectral_dni.data.apply(lambda g: integrate.trapz(spectral_dni.data.value, x=spectral_dni.data.index))
        dhi_scale = dhi / spectral_dhi.data.apply(lambda g: integrate.trapz(spectral_dhi.data.value, x=spectral_dhi.data.index))
        ghi_scale = ghi / spectral_ghi.data.apply(lambda g: integrate.trapz(spectral_ghi.data.value, x=spectral_ghi.data.index))
        # dni_scale = dni / (10*np.sum(spectral_dni[range(280, 4000, 10)]))
        # dhi_scale = dhi / (10*np.sum(spectral_dhi[range(280, 4000, 10)]))
        # ghi_scale = ghi / (10*np.sum(spectral_ghi[range(280, 2501, 10)]))
        spectral_dni.scale_values(dni_scale.value)
        spectral_dhi.scale_values(dhi_scale.value)
        spectral_ghi.scale_values(ghi_scale.value)
    
    # Write irradiance spectra
    #'''
    spectral_dni.to_file(dni_file)
    spectral_dhi.to_file(dhi_file)
    spectral_ghi.to_file(ghi_file)
    #'''
    
    # Generate/Load albedo
    alb_file = os.path.join(spectra_folder, "alb"+suffix)
    
    if material == 'Seasonal':
        MONTH = metdata.datetime[idx].month
        if 4 <= MONTH <= 7:
            material = 'Grass'
        else:
            material = 'DryGrass'
    spectral_alb = spectral_albedo_smarts(zen, azm, material, min_wavelength=300)
 
    # If specifed, scale the spectral albedo to have a mean value matching the
    # measured albedo.
    if scale_albedo:
        # option A
        denom = spectral_alb.data.value * spectral_ghi.data.value
        # option B
        #denom = spectral_alb.data

        # TODO:
        # Add test to
        if alb > 1 or alb == 0:
            print("albedo measured is an incorrect number, not scaling albedo generated")
        else:
            alb_scale = alb / denom.apply(lambda g: integrate.trapz(denom.values, x=spectral_alb.data.index))
            spectral_alb.scale_values(alb_scale.values)

    if scale_albedo_nonspectral_sim:
        sim_alb = np.sum(spectral_alb[range(280, 2501, 10)] * spectral_ghi[range(280, 2501, 10)])/np.sum(spectral_ghi[range(280, 2501, 10)])
        
        if alb > 1:
            print("albedo measured is an incorrect number, not scaling albedo generated")
        else:
            alb_scale = alb / sim_alb
            spectral_alb.scale_values(alb_scale)
            print(alb, sim_alb, alb_scale)
    
    # Write albedo file
    spectral_alb.to_file(alb_file)
    
    return (spectral_alb, spectral_dni, spectral_dhi)


"""Module containing specific irradiance modeling functions and tools.
More work probably needs to be done to improve the computation speed here, and
in theory a big part of ``perez_diffuse_luminance`` should be handled by pvlib
"""
import numpy as np
import pandas as pd
from pvlib import atmosphere, irradiance
from pvlib.tools import cosd, sind
import math
from Vendor.pvfactors.__init__ import PVFactorsError
from Vendor.pvfactors.config import \
    SIGMA, N_SIGMA, GAUSSIAN_DIAMETER_CIRCUMSOLAR, RADIUS_CIRCUMSOLAR


def perez_diffuse_luminance(timestamps, surface_tilt, surface_azimuth,
                            solar_zenith, solar_azimuth, dni, dhi):
    """Function used to calculate the luminance and the view factor terms from the
    Perez diffuse light transposition model, as implemented in the
    ``pvlib-python`` library.
    This function was custom made to allow the calculation of the circumsolar
    component on the back surface as well. Otherwise, the ``pvlib``
    implementation would ignore it.

    Parameters
    ----------
    timestamps : array-like
        simulation timestamps
    surface_tilt : array-like
        Surface tilt angles in decimal degrees.
        surface_tilt must be >=0 and <=180.
        The tilt angle is defined as degrees from horizontal
        (e.g. surface facing up = 0, surface facing horizon = 90)
    surface_azimuth : array-like
        The azimuth of the rotated panel,
        determined by projecting the vector normal to the panel's surface to
        the earth's surface [degrees].
    solar_zenith : array-like
        solar zenith angles
    solar_azimuth : array-like
        solar azimuth angles
    dni : array-like
        values for direct normal irradiance
    dhi : array-like
        values for diffuse horizontal irradiance

    Returns
    -------
    df_inputs : `pandas.DataFrame`
        Dataframe with the following columns:
        ['solar_zenith', 'solar_azimuth', 'surface_tilt', 'surface_azimuth',
        'dhi', 'dni', 'vf_horizon', 'vf_circumsolar', 'vf_isotropic',
        'luminance_horizon', 'luminance_circuqmsolar', 'luminance_isotropic',
        'poa_isotropic', 'poa_circumsolar', 'poa_horizon', 'poa_total_diffuse']

    """
    # Create a dataframe to help filtering on all arrays
    df_inputs = pd.DataFrame(
        {'surface_tilt': surface_tilt, 'surface_azimuth': surface_azimuth,
         'solar_zenith': solar_zenith, 'solar_azimuth': solar_azimuth,
         'dni': dni, 'dhi': dhi},
        index=pd.DatetimeIndex(timestamps))

    dni_et = irradiance.get_extra_radiation(df_inputs.index.dayofyear)
    am = atmosphere.get_relative_airmass(df_inputs.solar_zenith)

    # Need to treat the case when the sun is hitting the back surface of pvrow
    aoi_proj = irradiance.aoi_projection(
        df_inputs.surface_tilt, df_inputs.surface_azimuth,
        df_inputs.solar_zenith, df_inputs.solar_azimuth)
    sun_hitting_back_surface = ((aoi_proj < 0) &
                                (df_inputs.solar_zenith <= 90))
    df_inputs_back_surface = df_inputs.loc[sun_hitting_back_surface].copy()
    # Reverse the surface normal to switch to back-surface circumsolar calc
    df_inputs_back_surface.loc[:, 'surface_azimuth'] = (
        df_inputs_back_surface.loc[:, 'surface_azimuth'] - 180.)
    df_inputs_back_surface.loc[:, 'surface_azimuth'] = np.mod(
        df_inputs_back_surface.loc[:, 'surface_azimuth'], 360.
    )
    df_inputs_back_surface.loc[:, 'surface_tilt'] = (
        180. - df_inputs_back_surface.surface_tilt)

    if df_inputs_back_surface.shape[0] > 0:
        # Use recursion to calculate circumsolar luminance for back surface
        df_inputs_back_surface = perez_diffuse_luminance(
            *breakup_df_inputs(df_inputs_back_surface))

    # Calculate Perez diffuse components
    components = irradiance.perez(df_inputs.surface_tilt,
                                  df_inputs.surface_azimuth,
                                  df_inputs.dhi, df_inputs.dni,
                                  dni_et,
                                  df_inputs.solar_zenith,
                                  df_inputs.solar_azimuth,
                                  am,
                                  return_components=True)

    # Calculate Perez view factors:
    a = irradiance.aoi_projection(
        df_inputs.surface_tilt,
        df_inputs.surface_azimuth, df_inputs.solar_zenith,
        df_inputs.solar_azimuth)
    a = np.maximum(a, 0)
    b = cosd(df_inputs.solar_zenith)
    b = np.maximum(b, cosd(85))

    vf_perez = pd.DataFrame({
        'vf_horizon': sind(df_inputs.surface_tilt),
        'vf_circumsolar': a / b,
        'vf_isotropic': (1. + cosd(df_inputs.surface_tilt)) / 2.
    },
        index=df_inputs.index
    )

    # Calculate diffuse luminance
    luminance = pd.DataFrame(
        np.array([
            components['horizon'] / vf_perez['vf_horizon'],
            components['circumsolar'] / vf_perez['vf_circumsolar'],
            components['isotropic'] / vf_perez['vf_isotropic']
        ]).T,
        index=df_inputs.index,
        columns=['luminance_horizon', 'luminance_circumsolar',
                 'luminance_isotropic']
    )
    luminance.loc[components['sky_diffuse'] == 0, :] = 0.

    # Format components column names
    components = components.rename(columns={'isotropic': 'poa_isotropic',
                                            'circumsolar': 'poa_circumsolar',
                                            'horizon': 'poa_horizon'})

    df_inputs = pd.concat([df_inputs, components, vf_perez, luminance],
                          axis=1, join='outer')
    df_inputs = df_inputs.rename(columns={'sky_diffuse': 'poa_total_diffuse'})

    # Adjust the circumsolar luminance when it hits the back surface
    if df_inputs_back_surface.shape[0] > 0:
        df_inputs.loc[sun_hitting_back_surface, 'luminance_circumsolar'] = (
            df_inputs_back_surface.loc[:, 'luminance_circumsolar'])

    return df_inputs


def breakup_df_inputs(df_inputs):
    """Helper function: It is sometimes easier to provide a dataframe than a list
    of arrays: this function does the job of breaking up the dataframe into a
    list of expected 1-dim arrays

    Parameters
    ----------
    df_inputs : ``pandas.DataFrame``
        timestamp-indexed dataframe with following columns:
        'surface_azimuth', 'surface_tilt', 'solar_zenith', 'solar_azimuth',
        'dni', 'dhi'

    Returns
    -------
    all 1-dim ``numpy.ndarray``
        ``timestamps``, ``tilt_angles``, ``surface_azimuth``,
        ``solar_zenith``, ``solar_azimuth``, ``dni``, ``dhi``

    """
    timestamps = pd.to_datetime(df_inputs.index)
    surface_azimuth = df_inputs.surface_azimuth.values
    surface_tilt = df_inputs.surface_tilt.values
    solar_zenith = df_inputs.solar_zenith.values
    solar_azimuth = df_inputs.solar_azimuth.values
    dni = df_inputs.dni.values
    dhi = df_inputs.dhi.values

    return (timestamps, surface_tilt, surface_azimuth,
            solar_zenith, solar_azimuth, dni, dhi)


def calculate_circumsolar_shading(percentage_distance_covered,
                                  model='uniform_disk'):
    """Select the model to calculate circumsolar shading based on the current PV
    array condition.

    Parameters
    ----------
    percentage_distance_covered : float
        this represents how much of the
        circumsolar diameter is covered by the neighboring row [in %]
    model : str, optional
        name of the circumsolar shading model to use:
        'uniform_disk' and 'gaussian' are the two models currently available
        (Default value = 'uniform_disk')

    Returns
    -------
    float
        shading percentage of circumsolar area

    """
    if model == 'uniform_disk':
        perc_shading = uniform_circumsolar_disk_shading(
            percentage_distance_covered)

    elif model == 'gaussian':
        perc_shading = gaussian_shading(percentage_distance_covered)

    else:
        raise PVFactorsError(
            'calculate_circumsolar_shading: model does not exist: ' +
            '%s' % model)

    return perc_shading


def integral_default_gaussian(y, x):
    """Calculate the value of the integral from x to y of the erf function

    Parameters
    ----------
    y : float
        upper limit
    x : float
        lower limit

    Returns
    -------
    float
        Calculated value of integral

    """
    return 0.5 * (math.erf(x) - math.erf(y))


def gaussian_shading(percentage_distance_covered):
    """Calculate circumsolar shading by assuming that the irradiance profile on
    a 2D section of the circumsolar disk is Gaussian

    Parameters
    ----------
    percentage_distance_covered : float
        [in %], proportion of the
        circumsolar disk area covered by the neighboring pvrow

    Returns
    -------
    float
        percentage shading in terms of irradiance (using gaussian profile)

    """
    if percentage_distance_covered < 0.:
        perc_shading = 0.
    elif percentage_distance_covered > 100.:
        perc_shading = 100.
    else:
        y = - N_SIGMA * SIGMA
        x = (y +
             percentage_distance_covered / 100. *
             GAUSSIAN_DIAMETER_CIRCUMSOLAR)
        area = integral_default_gaussian(y, x)
        total_gaussian_area = integral_default_gaussian(- N_SIGMA * SIGMA,
                                                        N_SIGMA * SIGMA)
        perc_shading = area / total_gaussian_area * 100.

    return perc_shading


def gaussian(x, mu=0., sigma=1.):
    """Gaussian density function

    Parameters
    ----------
    x : float
        argument of function
    mu : float, optional
        mean of the gaussian (Default value = 0.)
    sigma : float, optional
        standard deviation of the gaussian (Default value = 1.)

    Returns
    -------
    float
        value of guassian function at point x

    """
    return (1. / (sigma * np.sqrt(2. * np.pi)) *
            np.exp(- 0.5 * np.power((x - mu) / sigma, 2)))


def uniform_circumsolar_disk_shading(percentage_distance_covered):
    """Calculate the percentage shading of circumsolar irradiance. This
    model considers circumsolar to be a disk, and calculates the
    percentage shaded based on the percentage "distance covered",
    which is how much of the disk's diameter is covered by the
    neighboring object.

    Parameters
    ----------
    percentage_distance_covered : float
        distance covered of circumsolar disk diameter
        [% - values from 0 to 100]

    Returns
    -------
    float
        value of guassian function at point x

    """

    # Define a circumsolar disk
    r_circumsolar = RADIUS_CIRCUMSOLAR
    d_circumsolar = 2 * r_circumsolar
    area_circumsolar = np.pi * r_circumsolar**2.

    # Calculate circumsolar on case by case
    distance_covered = percentage_distance_covered / 100. * d_circumsolar

    if distance_covered <= 0:
        # No shading of circumsolar
        percent_shading = 0.
    elif (distance_covered > 0.) & (distance_covered <= r_circumsolar):
        # Theta is the full circle sector angle (not half) used to
        # calculate circumsolar shading
        theta = 2 * np.arccos((r_circumsolar - distance_covered) /
                              r_circumsolar)  # rad
        area_circle_sector = 0.5 * r_circumsolar ** 2 * theta
        area_triangle_sector = 0.5 * r_circumsolar ** 2 * np.sin(theta)
        area_shaded = area_circle_sector - area_triangle_sector
        percent_shading = area_shaded / area_circumsolar * 100.

    elif (distance_covered > r_circumsolar) & (distance_covered <
                                               d_circumsolar):
        distance_uncovered = d_circumsolar - distance_covered

        # Theta is the full circle sector angle (not half) used to
        # calculate circumsolar shading
        theta = 2 * np.arccos((r_circumsolar - distance_uncovered) /
                              r_circumsolar)  # rad
        area_circle_sector = 0.5 * r_circumsolar ** 2 * theta
        area_triangle_sector = 0.5 * r_circumsolar ** 2 * np.sin(theta)
        area_not_shaded = area_circle_sector - area_triangle_sector
        percent_shading = (1. - area_not_shaded / area_circumsolar) * 100.

    else:
        # Total shading of circumsolar: distance_covered >= d_circumsolar
        percent_shading = 100.

    return percent_shading


def calculate_horizon_band_shading(shading_angle, horizon_band_angle):
    """Calculate the percentage shading of the horizon band; which is just the
    proportion of what's masked by the neighboring row since we assume
    uniform luminance values for the band

    Parameters
    ----------
    shading_angle : np.ndarray
        the elevation angle to use for shading
    horizon_band_angle : float
        the total elevation angle of the horizon band

    Returns
    -------
    float
        shading percentage of horizon band

    """
    percent_shading = np.where(shading_angle >= horizon_band_angle, 100., 0.)
    percent_shading = np.where(
        (shading_angle >= 0) & (shading_angle < horizon_band_angle),
        100. * shading_angle / horizon_band_angle, percent_shading)
    return percent_shading
